from builtins import range

import configparser
import logging
import os
import random


import numpy as np
import pandas as pd


from openfisca_core import periods
from openfisca_survey_manager.survey_collections import SurveyCollection
from openfisca_survey_manager.surveys import Survey
from openfisca_survey_manager import default_config_files_directory, openfisca_survey_manager_location


log = logging.getLogger(__name__)


def make_input_dataframe_by_entity(tax_benefit_system, nb_persons, nb_groups):
    """Generate a dictionnary of dataframes containing nb_persons persons spread in nb_groups groups.

    Args:
      tax_benefit_system(TaxBenefitSystem): the tax_benefit_system to use
      nb_persons(int): the number of persons in the system
      nb_groups(int): the number of collective entities in the system

    Returns:
      A dictionary whose keys are entities and values the corresponding data frames

      Example:

        >>> from openfisca_survey_manager.input_dataframe_generator import make_input_dataframe_by_entity
        >>> from openfisca_country_template import CountryTaxBenefitSystem
        >>> tbs = CountryTaxBenefitSystem()
        >>> input_dataframe_by_entity = make_input_dataframe_by_entity(tbs, 400, 100)
        >>> sorted(input_dataframe_by_entity['person'].columns.tolist())
        ['household_id', 'household_role', 'household_role_index', 'person_id']
        >>> sorted(input_dataframe_by_entity['household'].columns.tolist())
        []
    """
    input_dataframe_by_entity = dict()
    person_entity = [entity for entity in tax_benefit_system.entities if entity.is_person][0]
    person_id = np.arange(nb_persons)
    input_dataframe_by_entity = dict()
    input_dataframe_by_entity[person_entity.key] = pd.DataFrame({
        person_entity.key + '_id': person_id,
        })
    input_dataframe_by_entity[person_entity.key].set_index('person_id')
    #
    seed = 42
    random.seed(seed)
    adults = [0] + sorted(random.sample(range(1, nb_persons), nb_groups - 1))
    members_entity_id = np.empty(nb_persons, dtype = int)
    # A role index is an index that every person within an entity has.
    # For instance, the 'first_parent' has role index 0, the 'second_parent' 1, the first 'child' 2, the second 2, etc.
    members_role_index = np.empty(nb_persons, dtype = int)
    id_group = -1
    for id_person in range(nb_persons):
        if id_person in adults:
            id_group += 1
            role_index = 0
        else:
            role_index = 2
        members_role_index[id_person] = role_index
        members_entity_id[id_person] = id_group

    for entity in tax_benefit_system.entities:
        if entity.is_person:
            continue
        key = entity.key
        person_dataframe = input_dataframe_by_entity[person_entity.key]
        person_dataframe[key + '_id'] = members_entity_id
        person_dataframe[key + '_role_index'] = members_role_index
        person_dataframe[key + '_role'] = np.where(
            members_role_index == 0, entity.flattened_roles[0].key, entity.flattened_roles[-1].key)
        input_dataframe_by_entity[key] = pd.DataFrame({
            key + '_id': range(nb_groups)
            })
        input_dataframe_by_entity[key].set_index(key + '_id', inplace = True)

    return input_dataframe_by_entity


def random_data_generator(tax_benefit_system, nb_persons, nb_groups, variable_generators_by_period, collection = None):
    """
    Generate randomn values for some variables of a tax-benefit system and store them in a specified collection.

    Args:
        tax_benefit_system (TaxBenefitSystem): tax_benefit_system: the tax_benefit_system to use
        nb_persons (int): the number of persons in the data
        nb_groups (int): the number of collective entities in the data
        variable_generators_by_period (dict): parameters of the varaibles for every period
        collection (str, optional): collection where to store the input survey. Defaults to None.

    Returns:
        dict: The entities tables by period
    """
    initial_input_dataframe_by_entity = make_input_dataframe_by_entity(tax_benefit_system, nb_persons, nb_groups)
    table_by_entity_by_period = dict()
    for period, variable_generators in sorted(variable_generators_by_period.items()):
        input_dataframe_by_entity = initial_input_dataframe_by_entity.copy()
        table_by_entity_by_period[period] = table_by_entity = dict()
        for variable_generator in variable_generators:
            variable = variable_generator['variable']
            max_value = variable_generator['max_value']
            condition = variable_generator.get("condition", None)
            randomly_init_variable(
                tax_benefit_system = tax_benefit_system,
                input_dataframe_by_entity = input_dataframe_by_entity,
                variable_name = variable,
                max_value = max_value,
                condition = condition,
                )

        for entity, input_dataframe in input_dataframe_by_entity.items():
            if collection is not None:
                set_table_in_survey(input_dataframe, entity, period, collection, survey_name = 'input')

            table_by_entity[entity] = entity + '_' + str(period)

    return table_by_entity_by_period


def randomly_init_variable(tax_benefit_system, input_dataframe_by_entity, variable_name, max_value, condition = None, seed = None):
    """Initialises a variable with random values (from 0 to max_value).
        If a condition vector is provided, only set the value of persons or groups for which condition is True.

    Args:
      tax_benefit_system(TaxBenefitSystem): A tax benefit system
      input_dataframe_by_entity(dict): A dictionnary of entity dataframes
      variable_name: The name of the variable to initialize
      max_value: Maximum value of the variable
      condition: Boolean vector of obersvations to modify (Default value = None)
      seed: Random seed used whe ndrawing the values (Default value = None)

    Examples
        >>> from openfisca_survey_manager.input_dataframe_generator import make_input_dataframe_by_entity
        >>> from openfisca_country_template import CountryTaxBenefitSystem
        >>> tbs = CountryTaxBenefitSystem()
        >>> input_dataframe_by_entity = make_input_dataframe_by_entity(tbs, 400, 100)
        >>> randomly_init_variable(tbs, input_dataframe_by_entity, 'salary', max_value = 50000, condition = "household_role == 'first_parent'")  # Randomly set a salaire_net for all persons between 0 and 50000?
        >>> sorted(input_dataframe_by_entity['person'].columns.tolist())
        ['household_id', 'household_role', 'household_role_index', 'person_id', 'salary']
        >>> input_dataframe_by_entity['person'].salary.max() <= 50000
        True
        >>> len(input_dataframe_by_entity['person'].salary)
        400
        >>> randomly_init_variable(tbs, input_dataframe_by_entity, 'rent', max_value = 1000)
        >>> sorted(input_dataframe_by_entity['household'].columns.tolist())
        ['rent']
        >>> input_dataframe_by_entity['household'].rent.max() <= 1000
        True
        >>> input_dataframe_by_entity['household'].rent.max() >= 1
        True
        >>> len(input_dataframe_by_entity['household'].rent)
        100
    """

    variable = tax_benefit_system.variables[variable_name]
    entity = variable.entity

    if condition is None:
        condition = True
    else:
        condition = input_dataframe_by_entity[entity.key].eval(condition).values

    if seed is None:
        seed = 42
    np.random.seed(seed)
    count = len(input_dataframe_by_entity[entity.key])
    value = (np.random.rand(count) * max_value * condition).astype(variable.dtype)
    input_dataframe_by_entity[entity.key][variable_name] = value


def set_table_in_survey(input_dataframe, entity, period, collection, survey_name, survey_label = None,
        table_label = None, table_name = None, config_files_directory = default_config_files_directory):
    period = periods.period(period)
    if table_name is None:
        table_name = entity + '_' + str(period)
    if table_label is None:
        table_label = "Input data for entity {} at period {}".format(entity, period)
    try:
        survey_collection = SurveyCollection.load(collection = collection, config_files_directory=config_files_directory)
    except configparser.NoOptionError as e:
        log.warning(f"set_table_in_survey configparser.NoOptionError : {e}")
        survey_collection = SurveyCollection(name = collection)
    except configparser.NoSectionError as e:  # For tests
        log.warning(f"set_table_in_survey configparser.NoSectionError : {e}")
        data_dir = os.path.join(
            openfisca_survey_manager_location,
            'openfisca_survey_manager',
            'tests',
            'data_files',
            )
        survey_collection = SurveyCollection(
            name = collection,
            config_files_directory = data_dir,
            )

    try:
        survey = survey_collection.get_survey(survey_name)
    except AssertionError:
        log.info(f"Survey {survey_name} does not exist, it will be created.")
        survey = Survey(
            name = survey_name,
            label = survey_label or None,
            survey_collection = survey_collection,
            )

    if survey.hdf5_file_path is None:
        config = survey.survey_collection.config
        directory_path = config.get("data", "output_directory")
        if not os.path.isdir(directory_path):
            log.warning("{} who should be the HDF5 data directory does not exist: we create the directory".format(
                directory_path))
            os.makedirs(directory_path)
        survey.hdf5_file_path = os.path.join(directory_path, survey.name + '.h5')

    assert survey.hdf5_file_path is not None
    survey.insert_table(label = table_label, name = table_name, dataframe = input_dataframe)
    # If a survey with save name exist it will be overwritten
    survey_collection.surveys = [
        kept_survey for kept_survey in survey_collection.surveys if kept_survey.name != survey_name
        ]
    survey_collection.surveys.append(survey)
    collections_directory = survey_collection.config.get('collections', 'collections_directory')
    assert os.path.isdir(collections_directory), """{} who should be the collections' directory does not exist.
Fix the option collections_directory in the collections section of your config file.""".format(collections_directory)
    collection_json_path = os.path.join(collections_directory, "{}.json".format(collection))
    survey_collection.dump(json_file_path = collection_json_path)


def build_input_dataframe_from_test_case(survey_scenario, test_case_scenario_kwargs, period = None,
        computed_variables = None):
    if computed_variables is None:
        computed_variables = list()
    tax_benefit_system = survey_scenario.tax_benefit_system
    simulation = tax_benefit_system.new_scenario().init_single_entity(
        **test_case_scenario_kwargs
        ).new_simulation()
    array_by_variable = dict()
    period = periods.period(period)

    def compute_variable(variable):
        if variable not in tax_benefit_system.variables:
            return
        if period.unit == periods.YEAR:
            try:
                array_by_variable[variable] = simulation.calculate(variable, period = period)
            except Exception as e:
                log.debug(e)
                try:
                    array_by_variable[variable] = simulation.calculate_add(variable, period = period)
                except Exception as e:
                    log.debug(e)
                    array_by_variable[variable] = simulation.calculate(variable, period = period.first_month)
        elif period.unit == periods.MONTH:
            try:
                array_by_variable[variable] = simulation.calculate(variable, period = period)
            except ValueError as e:
                log.debug(e)
                array_by_variable[variable] = simulation.calculate(variable, period = period.this_year) / 12

    for scenario_key, value_by_variable in test_case_scenario_kwargs.items():
        if scenario_key == 'axes':
            variables = [test_case_scenario_kwargs['axes'][0][0]['name']]

        else:
            if value_by_variable is None:  # empty parent2 for example
                continue
            if not isinstance(value_by_variable, dict):  # enfants
                continue
            variables = list(value_by_variable.keys())

        for variable in variables:
            compute_variable(variable)

    for variable in computed_variables:
        compute_variable(variable)

    for entity in tax_benefit_system.entities:
        if entity.is_person:
            continue
        array_by_variable[
            survey_scenario.id_variable_by_entity_key[entity.key]
            ] = range(test_case_scenario_kwargs['axes'][0][0]['count'])

    input_data_frame = pd.DataFrame(array_by_variable)

    for entity in tax_benefit_system.entities:
        if entity.is_person:
            continue
        input_data_frame[
            survey_scenario.role_variable_by_entity_key[entity.key]
            ] = 0
    return input_data_frame



from openfisca_country_template.reforms.modify_social_security_taxation import modify_social_security_taxation
from openfisca_survey_manager.aggregates import AbstractAggregates
from openfisca_survey_manager.tests.test_scenario import create_randomly_initialized_survey_scenario


def test_aggregates():

    survey_scenario = create_randomly_initialized_survey_scenario(reform = modify_social_security_taxation)
    period = "2017-01"

    variables = [
        "social_security_contribution",
        "salary",
        ]
    aggregates = AbstractAggregates(survey_scenario=survey_scenario)
    aggregates.amount_unit = 1.0
    aggregates.beneficiaries_unit = 1.0
    aggregates.aggregate_variables = variables

    df = aggregates.compute_aggregates(reform = True, actual = False)

    for variable in variables:
        aggregate_before = survey_scenario.compute_aggregate(variable, period = period, use_baseline = True)
        aggregate_after = survey_scenario.compute_aggregate(variable, period = period)
        assert df.loc[variable, "baseline_amount"] == int(aggregate_before)
        assert df.loc[variable, "reform_amount"] == int(aggregate_after)



from openfisca_core.model_api import ETERNITY, Variable, Reform  # noqa analysis:ignore
from openfisca_country_template import CountryTaxBenefitSystem
from openfisca_country_template.entities import Person, Household


class Plugin(Reform):
    def apply(self):
        class person_weight(Variable):
            is_period_size_independent = True
            value_type = float
            entity = Person
            label = "Person weight"
            definition_period = ETERNITY

            def formula(person, period):
                return person.household('household_weight', period)

        class household_weight(Variable):
            is_period_size_independent = True
            value_type = float
            entity = Household
            label = "Household weight"
            definition_period = ETERNITY

        self.add_variable(person_weight)
        self.add_variable(household_weight)


tax_benefit_system = Plugin(CountryTaxBenefitSystem())



import pytest


from openfisca_core.tools import assert_near


from openfisca_survey_manager.tests.test_scenario import create_randomly_initialized_survey_scenario
from openfisca_survey_manager.calibration import Calibration


def test_calibration_variable_entity_is_weight_entity():
    survey_scenario = create_randomly_initialized_survey_scenario(collection=None)
    period = "2017-01"
    survey_scenario.period = period
    person_weight_before = survey_scenario.calculate_series("person_weight", period)
    # 'initial_rent_aggregate' is assigned to but never used initial_rent_aggregate = survey_scenario.compute_aggregate("rent", period = period)
    target_rent_aggregate = 200000

    survey_scenario.calibrate(
        period,
        target_margins_by_variable = {
            'rent': target_rent_aggregate,
            },
        parameters = {"method": "raking ratio"},
        )

    for _, simulation in survey_scenario.simulations.items():
        assert all(simulation.calibration.weight != simulation.calibration.initial_weight)

    assert_near(survey_scenario.compute_aggregate("rent", period = period), target_rent_aggregate)

    # See if propagation to derived weights is done well
    person_weight_after = survey_scenario.calculate_series("person_weight", period)
    assert (person_weight_after != person_weight_before).all()


def test_calibration_variable_entity_is_not_weight_entity():
    survey_scenario = create_randomly_initialized_survey_scenario(collection=None)
    period = "2017-01"
    survey_scenario.period = period
    target_rent_aggregate = 200000
    target_salary_aggregate = 1e7

    with pytest.raises(NotImplementedError):
        survey_scenario.calibrate(
            period,
            target_margins_by_variable = {
                'rent': target_rent_aggregate,
                'salary': target_salary_aggregate
                },
            parameters = {"method": "raking ratio"},
            )


def test_simulation_calibration_variable_entity_is_weight_entity():
    survey_scenario = create_randomly_initialized_survey_scenario(collection=None)
    period = "2017-01"
    survey_scenario.period = period
    simulation = list(survey_scenario.simulations.values())[0]
    person_weight_before = simulation.calculate("person_weight", period)

    # initial_rent_aggregate = simulation.compute_aggregate("rent", period = period)
    target_rent_aggregate = 200000

    calibration = Calibration(
        simulation,
        period = "2017-01",
        target_margins = {
            'rent': target_rent_aggregate,
            },
        target_entity_count = 300,
        parameters = {"method": "raking ratio"},
        )

    calibration.calibrate(inplace = True)
    assert all(calibration.weight != calibration.initial_weight)

    assert_near(simulation.compute_aggregate("rent", period = period), target_rent_aggregate)

    # See if propagation to derived weights is done well
    person_weight_after = simulation.calculate("person_weight", period)
    assert all(person_weight_after != person_weight_before)
    assert calibration.initial_entity_count != calibration.target_entity_count
    assert simulation.calculate("household_weight", period).sum() == calibration.target_entity_count



import logging


from openfisca_survey_manager.tests.test_scenario import create_randomly_initialized_survey_scenario


log = logging.getLogger(__name__)


def test_create_data_frame_by_entity():
    survey_scenario = create_randomly_initialized_survey_scenario()
    period = '2017-01'
    df_by_entity = survey_scenario.create_data_frame_by_entity(
        variables = ['salary', 'rent'],
        period = period,
        )
    salary = survey_scenario.calculate_variable('salary', period = period)
    rent = survey_scenario.calculate_variable('rent', period = period)
    for entity, df in df_by_entity.items():
        assert not df.empty, "{} dataframe is empty".format(entity)
    assert (df_by_entity['person']['salary'] == salary).all().all()
    assert (df_by_entity['household']['rent'] == rent).all().all()


def test_create_data_frame_by_entity_with_index():
    survey_scenario = create_randomly_initialized_survey_scenario()
    period = '2017-01'
    data_frame_by_entity = survey_scenario.create_data_frame_by_entity(
        variables = ['salary', 'rent', "person_id", "household_id"],
        period = period,
        index = True
        )
    for entity, input_dataframe in data_frame_by_entity.items():
        print(f"{entity} for {period}")  # noqa T201
        print(input_dataframe.columns)  # noqa T201
        if entity == "person":
            assert "person_id" in input_dataframe.columns.to_list()
        if entity == "household":
            assert "household_id" in input_dataframe.columns.to_list()



#! /usr/bin/env python


import collections
import os
import re

import logging
import pandas
import yaml


from .tables import Table


ident_re = re.compile(r"(?i)ident\d{2,4}$")  # noqa

log = logging.getLogger(__name__)


source_format_by_extension = dict(
    csv = 'csv',
    sas7bdat = "sas",
    dta = 'stata',
    Rdata = 'Rdata',
    spss = 'sav'
    )


class Survey(object):
    """An object to describe survey data"""
    hdf5_file_path = None
    informations = dict()
    label = None
    name = None
    tables = collections.OrderedDict()
    tables_index = dict()
    survey_collection = None

    def __init__(self, name = None, label = None, hdf5_file_path = None,
            survey_collection = None, **kwargs):
        assert name is not None, "A survey should have a name"
        self.name = name
        self.tables = dict()

        if label is not None:
            self.label = label

        if hdf5_file_path is not None:
            self.hdf5_file_path = hdf5_file_path

        if survey_collection is not None:
            self.survey_collection = survey_collection

        self.informations = kwargs

    def __repr__(self):
        header = """{} : survey data {}
Contains the following tables : \n""".format(self.name, self.label)
        tables = yaml.safe_dump(
            list(self.tables.keys()),
            default_flow_style = False)
        informations = yaml.safe_dump(self.informations, default_flow_style = False)
        return header + tables + informations

    @classmethod
    def create_from_json(cls, survey_json):
        self = cls(
            name = survey_json.get('name'),
            label = survey_json.get('label'),
            hdf5_file_path = survey_json.get('hdf5_file_path'),
            **survey_json.get('informations', dict())
            )
        self.tables = survey_json.get('tables')
        return self

    def dump(self):
        assert self.survey_collection is not None
        self.survey_collection.dump()

    def fill_hdf(self, source_format = None, tables = None, overwrite = True):
        assert self.survey_collection is not None
        assert isinstance(overwrite, bool) or isinstance(overwrite, list)
        survey = self
        if survey.hdf5_file_path is None:
            config = survey.survey_collection.config
            directory_path = config.get("data", "output_directory")
            if not os.path.isdir(directory_path):
                log.warn("{} who should be the HDF5 data directory does not exist: we create the directory".format(
                    directory_path))
                os.makedirs(directory_path)

            survey.hdf5_file_path = os.path.join(directory_path, survey.name + '.h5')
        if source_format is None:
            source_formats = ['csv', 'stata', 'sas', 'spss', 'Rdata']
        else:
            source_formats = [source_format]
        for source_format in source_formats:
            files = "{}_files".format(source_format)
            for data_file in survey.informations.get(files, []):
                path_name, extension = os.path.splitext(data_file)
                name = os.path.basename(path_name)
                if tables is None or name in tables:
                    table = Table(
                        label = name,
                        name = name,
                        source_format = source_format_by_extension[extension[1:]],
                        survey = survey,
                        )
                    table.fill_hdf(
                        data_file = data_file,
                        clean = True,
                        overwrite = overwrite if isinstance(overwrite, bool) else table.name in overwrite,
                        )
        self.dump()

    def find_tables(self, variable = None, tables = None, rename_ident = True):
        container_tables = []

        assert variable is not None

        if tables is None:
            tables = self.tables
        tables_index = self.tables_index
        for table in tables:
            if table not in tables_index:
                tables_index[table] = self.get_columns(table)
            if variable in tables_index[table]:
                container_tables.append(table)
        return container_tables

    def get_columns(self, table = None, rename_ident = True):
        assert table is not None
        store = pandas.HDFStore(self.hdf5_file_path, "r")
        if table in store:
            log.debug("Building columns index for table {}".format(table))
            data_frame = store[table]
            if rename_ident is True:
                for column_name in data_frame:
                    if ident_re.match(column_name) is not None:
                        data_frame.rename(columns = {column_name: "ident"}, inplace = True)
                        log.info("{} column have been replaced by ident".format(column_name))
                        break
            store.close()
            return list(data_frame.columns)
        else:
            log.info('table {} was not found in {}'.format(table, store.filename))
            store.close()
            return list()

    def get_value(self, variable, table, lowercase = False, ignorecase = False):
        """Get variable value from a survey table.

        Args:
          variable: variable to retrieve
          table(str): name of the table
          lowercase(bool, optional, optional): lowercase variable names, defaults to False
          ignorecase: ignore case of table name, defaults to False

        Returns:
          pd.DataFrame: dataframe containing the variable

        """
        return self.get_values([variable], table)

    def get_values(self, variables = None, table = None, lowercase = False, ignorecase = False, rename_ident = True):
        """Get variables values from a survey table.

        Args:
          variables(list, optional, optional): variables to retrieve, defaults to None (retrieve all variables)
          table(str, optional, optional): name of the table, defaults to None
          ignorecase: ignore case of table name, defaults to False
          lowercase(bool, optional, optional): lowercase variable names, defaults to False
          rename_ident(bool, optional, optional): rename ident+yr (e.g. ident08) into ident, defaults to True

        Returns:
          pd.DataFrame: dataframe containing the variables

        Raises:
          Exception:

        """
        assert self.hdf5_file_path is not None
        assert os.path.exists(self.hdf5_file_path), '{} is not a valid path. This could happen because your data were not builded yet. Please consider using a rebuild option in your code.'.format(
            self.hdf5_file_path)
        store = pandas.HDFStore(self.hdf5_file_path, "r")

        if ignorecase:
            keys = store.keys()
            eligible_tables = []
            for string in keys:
                match = re.findall(table, string, re.IGNORECASE)
                if match:
                    eligible_tables.append(match[0])
            if len(eligible_tables) > 1:
                raise ValueError(f"{table} is ambiguious since the following tables are available: {eligible_tables}")
            elif len(eligible_tables) == 0:
                raise ValueError(f"No eligible available table in {keys}")
            else:
                table = eligible_tables[0]
        try:
            df = store.select(table)
        except KeyError:
            log.error(f'No table {table} in the file {self.hdf5_file_path}')
            log.error(f'This could happen because your data were not builded yet. Available tables are: {store.keys()}')
            store.close()
            raise

        store.close()
        if lowercase:
            columns = dict((column_name, column_name.lower()) for column_name in df)
            df.rename(columns = columns, inplace = True)

        if rename_ident is True:
            for column_name in df:
                if ident_re.match(str(column_name)) is not None:
                    df.rename(columns = {column_name: "ident"}, inplace = True)
                    log.info("{} column have been replaced by ident".format(column_name))
                    break

        if variables is None:
            return df
        else:
            diff = set(variables) - set(df.columns)
            if diff:
                raise Exception("The following variable(s) {} are missing".format(diff))
            variables = list(set(variables).intersection(df.columns))
            df = df[variables]
            return df

    def insert_table(self, label = None, name = None, **kwargs):
        """Inserts a table in the Survey object"""

        data_frame = kwargs.pop('data_frame', None)
        variables = kwargs.pop('variables', None)
        if data_frame is None:
            data_frame = kwargs.pop('dataframe', None)

        to_hdf_kwargs = kwargs.pop('to_hdf_kwargs', dict())
        if data_frame is not None:
            assert isinstance(data_frame, pandas.DataFrame)
            if variables is not None:
                assert set(variables) < set(data_frame.columns)
            else:
                variables = list(data_frame.columns)

        if data_frame is not None:
            if label is None:
                label = name
            table = Table(label = label, name = name, survey = self, variables = variables)
            assert table.survey.hdf5_file_path is not None
            log.debug("Saving table {} in {}".format(name, table.survey.hdf5_file_path))
            table.save_data_frame(data_frame, **to_hdf_kwargs)

        if name not in self.tables:
            self.tables[name] = dict()
        for key, val in kwargs.items():
            self.tables[name][key] = val

    def to_json(self):
        self_json = collections.OrderedDict((
            ))
        self_json['hdf5_file_path'] = self.hdf5_file_path
        self_json['label'] = self.label
        self_json['name'] = self.name
        self_json['tables'] = self.tables
        self_json['informations'] = collections.OrderedDict(sorted(self.informations.items()))
        return self_json



"""Monkey-patch openfisca_core.simulations.Simulation to work with pandas."""

import logging
import numpy as np
import pandas as pd
import re
from typing import Callable, Dict, List, Optional, Union


import humanize


from openfisca_core import periods
from openfisca_core.memory_config import MemoryConfig
from openfisca_core.indexed_enums import Enum
from openfisca_core.periods import ETERNITY, MONTH, YEAR
from openfisca_core.types import Array, Entity, Period, TaxBenefitSystem
from openfisca_core.simulations import Simulation
from openfisca_survey_manager.simulation_builder import diagnose_variable_mismatch, SimulationBuilder
from openfisca_survey_manager.survey_collections import SurveyCollection


from openfisca_survey_manager.statshelpers import mark_weighted_percentiles
from openfisca_survey_manager.utils import do_nothing, load_table


log = logging.getLogger(__name__)


# Helpers

def assert_variables_in_same_entity(tax_benefit_system: TaxBenefitSystem, variables: List):
    """
    Assert that variables are in the same entity.

    Args:
        tax_benefit_system (TaxBenefitSystem): Host tax benefit system
        variables (List): Variables supposed to belong to the same entity

    Returns:
        str: Common entity of the variables
    """
    entity = None
    for variable_name in variables:
        variable = tax_benefit_system.variables.get(variable_name)
        assert variable
        if entity is None:
            entity = variable.entity
        assert variable.entity == entity, "{} are not from the same entity: {} doesn't belong to {}".format(
            variables, variable_name, entity.key)
    return entity.key


def get_words(text: str):
    return re.compile('[A-Za-z_]+').findall(text)


# Main functions

def adaptative_calculate_variable(simulation: Simulation, variable: str, period: Optional[Union[int, str, Period]]) -> Array:
    """
    Calculate variable by adpating it definition period to the target period.

    Args:
        simulation (Simulation): Simulation to suse
        variable (str): Variable to be computed
        period (Optional[Union[int, str, Period]]): Target period

    Returns:
        Array: Values of the variable on the target period
    """
    if not isinstance(period, periods.Period):
        period = periods.period(str(period))

    tax_benefit_system = simulation.tax_benefit_system
    assert tax_benefit_system is not None

    assert variable in tax_benefit_system.variables, "{variable} is not a valid variable"
    period_size_independent = tax_benefit_system.get_variable(variable).is_period_size_independent
    definition_period = tax_benefit_system.get_variable(variable).definition_period

    if period_size_independent is False and definition_period != 'eternity':
        values = simulation.calculate_add(variable, period = period)
    elif period_size_independent is True and definition_period == 'month' and period.size_in_months > 1:
        values = simulation.calculate(variable, period = period.first_month)
    elif period_size_independent is True and definition_period == 'month' and period.size_in_months == 1:
        values = simulation.calculate(variable, period = period)
    elif period_size_independent is True and definition_period == 'year' and period.size_in_months > 12:
        values = simulation.calculate(variable, period = period.start.offset('first-of', 'year').period('year'))
    elif period_size_independent is True and definition_period == 'year' and period.size_in_months == 12:
        values = simulation.calculate(variable, period = period)
    elif period_size_independent is True and definition_period == 'year':
        values = simulation.calculate(variable, period = period.this_year)
    elif definition_period == 'eternity':
        values = simulation.calculate(variable, period = period)
    else:
        values = None
    assert values is not None, f"Unspecified calculation period for variable {variable}"

    return values


def compute_aggregate(simulation: Simulation, variable: str = None, aggfunc: str = 'sum', filter_by: str = None, period: Optional[Union[int, str, Period]] = None,
        missing_variable_default_value = np.nan, weighted: bool = True, alternative_weights: Optional[Union[str, int, float, Array]] = None,
        filtering_variable_by_entity: Dict = None) -> Optional[Union[None, float]]:
    """
    Compute aggregate of a variable.

    Args:
        simulation (Simulation): Simulation to use for the computation
        variable (str, optional): Variable to aggregate. Defaults to None.
        aggfunc (str, optional): Aggregation function. Defaults to 'sum'.
        filter_by (str, optional): Filter variable or expression to use. Defaults to None.
        period (Optional[Union[int, str, Period]], optional): Period. Defaults to None.
        missing_variable_default_value (optional): Value to use for missing values. Defaults to np.nan.
        weighted (bool, optional): Whether to weight the variable or not. Defaults to True.
        alternative_weights (Optional[Union[str, int, float, Array]], optional): Alternative weigh to use. Defaults to None.
        filtering_variable_by_entity (Dict, optional): Filtering variable by entity. Defaults to None.

    Returns:
        float: Aggregate
    """
    weight_variable_by_entity = simulation.weight_variable_by_entity
    tax_benefit_system = simulation.tax_benefit_system

    if period is None:
        period = simulation.period

    assert variable in tax_benefit_system.variables, f"{variable} is not a variable of the tax benefit system"
    entity_key = tax_benefit_system.variables[variable].entity.key

    if filter_by is None and filtering_variable_by_entity is not None:
        filter_by_variable = filtering_variable_by_entity.get(entity_key)

    if filter_by:
        filter_by_variable = get_words(filter_by)[0]
        assert filter_by_variable in tax_benefit_system.variables, f"{filter_by_variable} is not a variable of the tax benefit system"
        entity_key = tax_benefit_system.variables[variable].entity.key
        filter_by_entity_key = tax_benefit_system.variables[filter_by_variable].entity.key
        assert filter_by_entity_key == entity_key, (
            f"You tried to compute agregates for variable '{variable}', of entity {entity_key}"
            f" filtering by variable '{filter_by_variable}', of entity {filter_by_entity_key}. This is not possible."
            f" Please choose a filter-by variable of same entity as '{variable}'."
            )

    expressions = []
    if filter_by is not None:
        if filter_by in tax_benefit_system.variables:
            filter_entity_key = tax_benefit_system.variables.get(filter_by).entity.key
            assert filter_entity_key == entity_key, (
                "You tried to compute agregates for variable '{0}', of entity {1}"
                " filtering by variable '{2}', of entity {3}. This is not possible."
                " Please choose a filter-by variable of same entity as '{0}'.".format(
                    variable, entity_key, filter_by_variable, filter_by_entity_key
                    )
                )
        else:
            filter_entity_key = assert_variables_in_same_entity(tax_benefit_system, get_words(filter_by))
            expressions.extend([filter_by])
            assert filter_entity_key == entity_key
    else:
        filter_dummy = np.array(1.0)

    uniform_weight = np.array(1.0)
    weight_variable = None
    if weighted:
        if alternative_weights:
            if isinstance(alternative_weights, str):
                assert alternative_weights in tax_benefit_system.variables, \
                    f"{alternative_weights} is not a valid variable of the tax benefit system"
                weight_variable = alternative_weights

            elif (type(alternative_weights) is int) or (type(alternative_weights) is float):
                weight_variable = None
                uniform_weight = float(alternative_weights)

        else:
            if weight_variable_by_entity:
                weight_variable = weight_variable_by_entity[entity_key]

            else:
                log.warn('There is no weight variable for entity {} nor alternative weights. Switch to unweighted'.format(entity_key))

    if variable in simulation.tax_benefit_system.variables:
        value = simulation.adaptative_calculate_variable(variable = variable, period = period)
    else:
        log.debug("Variable {} not found. Assiging {}".format(variable, missing_variable_default_value))
        return missing_variable_default_value

    weight = (
        simulation.calculate(weight_variable, period = period).astype(float)
        if weight_variable else uniform_weight
        )
    if weight_variable:
        assert any(weight != 0), "Weights shall not be all zeroes"
    else:
        assert uniform_weight != 0

    if filter_by is not None:
        expression_data_frame = simulation.create_data_frame_by_entity(
            variables = get_words(filter_by),
            period = period,
            index = False
            )[entity_key]
        for expression in expressions:
            expression_data_frame[expression] = expression_data_frame.eval(expression)

        filter_dummy = expression_data_frame[filter_by]
    else:
        filter_dummy = 1.0

    if aggfunc == 'sum':
        aggregate = (value * weight * filter_dummy).sum()
    elif aggfunc == 'mean':
        aggregate = (value * weight * filter_dummy).sum() / (weight * filter_dummy).sum()
    elif aggfunc == 'count':
        aggregate = (weight * filter_dummy).sum()
    elif aggfunc == 'count_non_zero':
        aggregate = (weight * (value != 0) * filter_dummy).sum()
    else:
        aggregate = None

    return aggregate


def compute_quantiles(simulation: Simulation, variable: str, nquantiles: int = None,
        period: Optional[Union[int, str, Period]] = None, filter_by = None, weighted: bool = True,
        alternative_weights = None, filtering_variable_by_entity = None) -> List[float]:
    """
    Compute quantiles of a variable.

    Args:
        simulation (Simulation, optional): Simulation to be used. Defaults to None.
        variable (str, optional): Variable which quantiles are computed. Defaults to None.
        nquantiles (int, optional): Number of quantiles. Defaults to None.
        period (Optional[Union[int, str, Period]], optional): Period. Defaults to None.
        missing_variable_default_value (optional): Value to use for missing values. Defaults to np.nan.
        weighted (bool, optional): Whether to weight the variable or not. Defaults to True.
        alternative_weights (Optional[Union[str, int, float, Array]], optional): Alternative weigh to use. Defaults to None.
        filtering_variable_by_entity (Dict, optional): Filtering variable by entity. Defaults to None.

    Returns:
       List(float): The quantiles values
    """
    weight_variable_by_entity = simulation.weight_variable_by_entity
    weight_variable = None
    entity_key = simulation.tax_benefit_system.variables[variable].entity.key
    if weight_variable_by_entity:
        weight_variable = weight_variable_by_entity[entity_key]

    variable_values = simulation.adaptative_calculate_variable(variable, period)
    if weighted:
        assert (alternative_weights is not None) or (weight_variable is not None)
        weight = (
            alternative_weights
            if alternative_weights is not None
            else simulation.calculate(weight_variable, period)
            )
    else:
        weight = np.ones(len(variable_values))

    if filtering_variable_by_entity is not None:
        if filter_by is None:
            filter_by = filtering_variable_by_entity.get(entity_key)

    if filter_by is not None:
        filter_entity_key = simulation.tax_benefit_system.variables.get(filter_by).entity.key
        assert filter_entity_key == entity_key
        filter_dummy = simulation.calculate(filter_by, period = period).astype(bool)

        variable_values = variable_values[filter_dummy].copy()
        weight = weight[filter_dummy].copy()

    labels = np.arange(1, nquantiles + 1)
    method = 2
    _, values = mark_weighted_percentiles(variable_values, labels, weight, method, return_quantiles = True)
    return values


def compute_pivot_table(simulation: Simulation = None, baseline_simulation: Simulation = None, aggfunc = 'mean',
        columns: Optional[List[str]] = None, difference: bool = False, filter_by = None, index: Optional[List[str]] = None,
        period: Optional[Union[int, str, Period]] = None, use_baseline_for_columns: bool = None, values: Optional[List[str]] = None,
        missing_variable_default_value = np.nan, concat_axis: Optional[int] = None, weighted: bool = True, alternative_weights = None,
        filtering_variable_by_entity = None):
    """
    Compute pivot table.

    Args:
        simulation (Simulation, optional): Main simulation. Defaults to None.
        baseline_simulation (Simulation, optional): Baseline simulation. Defaults to None.
        aggfunc (str, optional): Aggregation function. Defaults to 'mean'.
        columns (List[str], optional): Variables to use in columns. Defaults to None.
        difference (bool, optional): Whether to compute the difference with baseline. Defaults to False.
        filter_by (str, optional): Filter variable or expression to use. Defaults to None.
        index (List[str], optional): _description_. Defaults to None.
        period (Optional[Union[int, str, Period]], optional): Period. Defaults to None.
        use_baseline_for_columns (bool, optional): _description_. Defaults to None.
        values (List[str], optional): _description_. Defaults to None.
        missing_variable_default_value (optional): _description_. Defaults to np.nan.
        concat_axis (int, optional): _description_. Defaults to None.
        weighted (bool, optional): Whether to weight the variable or not. Defaults to True.
        alternative_weights (Optional[Union[str, int, float, Array]], optional): Alternative weigh to use. Defaults to None.
        filtering_variable_by_entity (Dict, optional): Filtering variable by entity. Defaults to None.

    Returns:
        _type_: _description_
    """
    weight_variable_by_entity = simulation.weight_variable_by_entity

    admissible_aggfuncs = ['max', 'mean', 'min', 'sum', 'count', 'sum_abs']
    assert aggfunc in admissible_aggfuncs
    assert columns or index or values

    if baseline_simulation is not None:
        tax_benefit_system = baseline_simulation.tax_benefit_system
    else:
        tax_benefit_system = simulation.tax_benefit_system

    assert period is not None

    if isinstance(columns, str):
        columns = [columns]
    elif columns is None:
        columns = []
    assert isinstance(columns, list)

    if isinstance(index, str):
        index = [index]
    elif index is None:
        index = []
    assert isinstance(index, list)

    if isinstance(values, str):
        values = [values]
    elif values is None:
        values = []
    assert isinstance(values, list)

    entity_key = None
    for axe in [columns, index, values]:
        if (len(axe) != 0) and (entity_key is None):
            entity_key = tax_benefit_system.variables[axe[0]].entity.key
            continue

    if filter_by is None and filtering_variable_by_entity is not None:
        filter_by = filtering_variable_by_entity.get(entity_key)

    variables = set(index + columns)

    # Select the entity weight corresponding to the variables that will provide values
    uniform_weight = 1.0
    weight_variable = None
    if weighted:
        if alternative_weights:
            if isinstance(alternative_weights, str):
                assert alternative_weights in tax_benefit_system.variables, \
                    f"{alternative_weights} is not a valid variable of the tax benefit system"
                weight_variable = alternative_weights

            elif (type(alternative_weights) is int) or (type(alternative_weights) is float):
                weight_variable = None
                uniform_weight = float(alternative_weights)

        else:
            if weight_variable_by_entity:
                weight_variable = weight_variable_by_entity[entity_key]
                variables.add(weight_variable)

            else:
                log.warn('There is no weight variable for entity {} nor alternative weights. Switch to unweighted'.format(entity_key))

    expressions = []
    if filter_by is not None:
        if filter_by in tax_benefit_system.variables:
            variables.add(filter_by)
            filter_entity_key = tax_benefit_system.variables.get(filter_by).entity.key
            assert filter_entity_key == entity_key
        else:
            filter_entity_key = assert_variables_in_same_entity(tax_benefit_system, get_words(filter_by))
            expressions.extend([filter_by])
            assert filter_entity_key == entity_key
    else:
        filter_dummy = np.array(1.0)

    for expression in expressions:
        expression_variables = get_words(expression)
        expression_entity_key = assert_variables_in_same_entity(tax_benefit_system, expression_variables)
        assert expression_entity_key == entity_key
        for variable in expression_variables:
            variables.add(variable)

    for variable in variables | set(values):
        if variable in tax_benefit_system.variables:
            assert tax_benefit_system.variables[variable].entity.key == entity_key, \
                'The variable {} does not belong to entity {}'.format(
                    variable,
                    entity_key,
                    )

    if difference:
        assert simulation is not None and baseline_simulation is not None
        reform_data_frame = simulation.create_data_frame_by_entity(
            values, period = period, index = False
            )[entity_key].fillna(missing_variable_default_value)
        baseline_data_frame = baseline_simulation.create_data_frame_by_entity(
            values, period = period, index = False
            )[entity_key].fillna(missing_variable_default_value)
        for value_variable in values:
            if value_variable not in baseline_data_frame:
                baseline_data_frame[value_variable] = missing_variable_default_value
            if value_variable not in reform_data_frame:
                reform_data_frame[value_variable] = missing_variable_default_value

        data_frame = reform_data_frame - baseline_data_frame

    else:
        if values:
            data_frame = (
                simulation.create_data_frame_by_entity(
                    values, period = period, index = False)[entity_key]
                )
            for value_variable in values:
                if value_variable not in data_frame:
                    data_frame[value_variable] = missing_variable_default_value
        else:
            data_frame = None

    use_baseline_data = difference or use_baseline_for_columns

    # use baseline if explicited or when computing difference
    if use_baseline_data:
        baseline_vars_data_frame = baseline_simulation.create_data_frame_by_entity(
            variables = variables,
            period = period,
            index = False
            )[entity_key]
    else:
        baseline_vars_data_frame = simulation.create_data_frame_by_entity(
            variables = variables,
            period = period,
            index = False
            )[entity_key]

    for expression in expressions:
        baseline_vars_data_frame[expression] = baseline_vars_data_frame.eval(expression)
    if filter_by is not None:
        filter_dummy = baseline_vars_data_frame[filter_by]
    if weight_variable is None:
        weight_variable = 'weight'
        baseline_vars_data_frame[weight_variable] = uniform_weight
    baseline_vars_data_frame[weight_variable] = baseline_vars_data_frame[weight_variable] * filter_dummy
    # We drop variables that are in values from baseline_vars_data_frame
    dropped_columns = [
        column for column in baseline_vars_data_frame.columns if column in values
        ]
    baseline_vars_data_frame.drop(columns = dropped_columns, inplace = True)

    data_frame = pd.concat(
        [baseline_vars_data_frame, data_frame],
        axis = 1,
        )

    if values:
        data_frame_by_value = dict()
        for value in values:
            if aggfunc in ['mean', 'sum', 'sum_abs', 'count']:
                data_frame[value] = (
                    data_frame[value] * data_frame[weight_variable]
                    if aggfunc != 'sum_abs'
                    else data_frame[value].abs() * data_frame[weight_variable]
                    )
                data_frame[value].fillna(missing_variable_default_value, inplace = True)
                pivot_sum = data_frame.pivot_table(index = index, columns = columns, values = value, aggfunc = 'sum')
                pivot_mass = data_frame.pivot_table(index = index, columns = columns, values = weight_variable, aggfunc = 'sum')
                if aggfunc == 'mean':
                    try:  # Deal with a pivot_table pandas bug https://github.com/pandas-dev/pandas/issues/17038
                        result = (pivot_sum / pivot_mass.loc[weight_variable])
                    except KeyError:
                        result = (pivot_sum / pivot_mass)
                elif aggfunc in ['sum', 'sum_abs']:
                    result = pivot_sum
                elif aggfunc == 'count':
                    result = pivot_mass.rename(columns = {weight_variable: value}, index = {weight_variable: value})

            elif aggfunc in ["min", "max"]:
                data_frame[value].fillna(missing_variable_default_value, inplace = True)
                result = data_frame.pivot_table(index = index, columns = columns, values = value, aggfunc = aggfunc)

            data_frame_by_value[value] = result

        if len(list(data_frame_by_value.keys())) > 1:
            if concat_axis is None:
                return data_frame_by_value
            else:
                assert concat_axis in [0, 1]
                return pd.concat(data_frame_by_value.values(), axis = concat_axis)
        else:
            return next(iter(data_frame_by_value.values()))

    else:
        assert aggfunc == 'count', "Can only use count for aggfunc if no values"
        return data_frame.pivot_table(index = index, columns = columns, values = weight_variable, aggfunc = 'sum')


def create_data_frame_by_entity(simulation: Simulation, variables: Optional[List] = None, expressions: Optional[List[str]] = None,
        filter_by = None, index: bool = False, period: Optional[Union[int, str, Period]] = None,
        merge: bool = False) -> Union[pd.DataFrame, Dict]:
    """
    Create dataframe(s) of variables for the whole selected population.

    Args:
        simulation (Simulation): Simulation to use.
        variables (Optional[List], optional): Variables to retrieve, None means all. Defaults to None.
        expressions (Optional[List[str]], optional): _description_. Defaults to None.
        filter_by (str, optional): Filter variable or expression to use. Defaults to None.
        index (bool, optional): Whether to use index (id) variables. Defaults to False.
        period (Optional[Union[int, str, Period]], optional): Period of the computation. Defaults to None.
        merge (bool, optional): Wheter to merge the datafrales into one. Defaults to False.

    Returns:
        pd.DataFrame of Dict: Dataframe(s) with the variables values
    """
    assert simulation is not None
    id_variable_by_entity_key = simulation.id_variable_by_entity_key
    tax_benefit_system = simulation.tax_benefit_system
    assert tax_benefit_system is not None

    if period is None:
        period = simulation.period

    assert variables or index or expressions or filter_by

    if merge:
        index = True
    if expressions is None:
        expressions = []

    if filter_by is not None:
        if filter_by in tax_benefit_system.variables:
            variables.append(filter_by)
            filter_entity_key = tax_benefit_system.variables.get(filter_by).entity.key
        else:
            filter_entity_key = assert_variables_in_same_entity(tax_benefit_system, get_words(filter_by))
            expressions.append(filter_by)

    expressions_by_entity_key = dict()
    for expression in expressions:
        expression_variables = get_words(expression)
        entity_key = assert_variables_in_same_entity(tax_benefit_system, expression_variables)
        if entity_key in expressions_by_entity_key:
            expressions_by_entity_key[entity_key].append(expression)
        else:
            expressions_by_entity_key[entity_key] = [expression]
        variables += expression_variables

    variables = set(variables)

    missing_variables = set(variables).difference(set(tax_benefit_system.variables.keys()))
    if missing_variables:
        log.info("These variables aren't part of the tax-benefit system: {}".format(missing_variables))
        log.info("These variables are ignored: {}".format(missing_variables))

    columns_to_fetch = [
        tax_benefit_system.variables.get(variable_name) for variable_name in variables
        if tax_benefit_system.variables.get(variable_name) is not None
        ]

    assert len(columns_to_fetch) >= 1, "None of the requested variables {} are in the tax-benefit-system {}".format(
        variables, list(tax_benefit_system.variables.keys()))

    assert simulation is not None

    openfisca_data_frame_by_entity_key = dict()
    non_person_entities = list()

    for entity in tax_benefit_system.entities:
        entity_key = entity.key
        column_names = [
            column.name for column in columns_to_fetch
            if column.entity.key == entity_key
            ]
        openfisca_data_frame_by_entity_key[entity_key] = pd.DataFrame(
            dict(
                (column_name, simulation.adaptative_calculate_variable(column_name, period = period))
                for column_name in column_names
                )
            )
        if entity.is_person:
            person_entity = entity
        else:
            non_person_entities.append(entity)

    if index:
        person_data_frame = openfisca_data_frame_by_entity_key.get(person_entity.key)
        person_data_frame.index.name = id_variable_by_entity_key.get("person", "person_id")
        if person_data_frame is None:
            person_data_frame = pd.DataFrame()
        for entity in non_person_entities:
            entity_key_id = id_variable_by_entity_key[entity.key]
            person_data_frame[
                entity_key_id
                ] = simulation.populations[entity.key].members_entity_id
            flattened_roles = entity.flattened_roles
            index_by_role = dict(
                (flattened_roles[index], index)
                for index in range(len(flattened_roles))
                )
            person_data_frame[
                "{}_{}".format(entity.key, 'role')
                ] = pd.Series(simulation.populations[entity.key].members_role).map(index_by_role)
            person_data_frame[
                "{}_{}".format(entity.key, 'position')
                ] = simulation.populations[entity.key].members_position

            # Set index names as entity_id
            openfisca_data_frame_by_entity_key[entity.key].index.name = entity_key_id
            openfisca_data_frame_by_entity_key[entity.key].reset_index(inplace=True)
        person_data_frame.reset_index(inplace=True)

    for entity_key, expressions in expressions_by_entity_key.items():
        data_frame = openfisca_data_frame_by_entity_key[entity_key]
        for expression in expressions:
            data_frame[expression] = data_frame.eval(expression)

    if filter_by is not None:
        openfisca_data_frame_by_entity_key[filter_entity_key] = (
            openfisca_data_frame_by_entity_key[filter_entity_key].loc[
                openfisca_data_frame_by_entity_key[filter_entity_key][filter_by]
                ].copy()
            )

    if not merge:
        return openfisca_data_frame_by_entity_key
    else:
        for entity_key, openfisca_data_frame in openfisca_data_frame_by_entity_key.items():
            if entity_key != person_entity.key:
                entity_key_id = id_variable_by_entity_key[entity.key]
                if len(openfisca_data_frame) > 0:
                    person_data_frame = person_data_frame.merge(
                        openfisca_data_frame.reset_index(),
                        left_on = entity_key_id,
                        right_on = entity_key_id,
                        )
        return person_data_frame


class SecretViolationError(Exception):
    """Raised if the result of the simulation do not comform with regulators rules."""

    pass


def compute_winners_loosers(
        simulation: Simulation,
        baseline_simulation: Simulation,
        variable: str,
        filter_by: Optional[str] = None,
        period: Optional[Union[int, str, Period]] = None,
        absolute_minimal_detected_variation: float = 0,
        relative_minimal_detected_variation: float = .01,
        observations_threshold: int = None,
        weighted: bool = True,
        alternative_weights = None,
        filtering_variable_by_entity = None,
        ) -> Dict[str, int]:
    """
    Compute the number of winners and loosers for a given variable.

    Args:
        simulation (_type_): The main simulation.
        baseline_simulation (_type_): The baseline simulation
        variable (str): The variable to use.
        filter_by (str, optional): The variable or expression to be used as a filter. Defaults to None.
        period (Optional[Union[int, str, Period]], optional): The period of the simulation. Defaults to None.
        absolute_minimal_detected_variation (float, optional): Absolute minimal variation to be detected, in ratio. Ie 0.5 means 5% of variation wont be counted..
        relative_minimal_detected_variation (float, optional): Relative minimal variation to be detected, in ratio. Defaults to .01.
        observations_threshold (int, optional): Number of observations needed to avoid a statistical secret violation. Defaults to None.
        weighted (bool, optional): Whether to use weights. Defaults to True.
        alternative_weights (Optional[Union[str, int, float, Array]], optional): Alternative weigh to use. Defaults to None.
        filtering_variable_by_entity (_type_, optional): The variable to be used as a filter for each entity. Defaults to None.

    Raises:
        SecretViolationError: Raised when statistical secret is violated.

    Returns:
        Dict[str, int]: Statistics about winners and loosers between the main simulation and the baseline.
    """
    weight_variable_by_entity = simulation.weight_variable_by_entity
    entity_key = baseline_simulation.tax_benefit_system.variables[variable].entity.key

    # Get the results of the simulation
    after = simulation.adaptative_calculate_variable(variable, period = period)
    before = baseline_simulation.adaptative_calculate_variable(variable, period = period)

    # Filter if needed
    if filtering_variable_by_entity is not None:
        if filter_by is None:
            filter_by = filtering_variable_by_entity.get(entity_key)

    if filter_by is not None:
        filter_entity_key = baseline_simulation.tax_benefit_system.variables.get(filter_by).entity.key
        assert filter_entity_key == entity_key
        filter_dummy = baseline_simulation.calculate(filter_by, period = period).astype(bool)

        after = after[filter_dummy].copy()
        before = before[filter_dummy].copy()

    # Define weights
    weight = np.ones(len(after))
    if weighted:
        if alternative_weights is not None:
            weight = alternative_weights
        elif weight_variable_by_entity is not None:
            weight_variable = weight_variable_by_entity[entity_key]
            weight = baseline_simulation.calculate(weight_variable, period = period)
        else:
            log.warn('There is no weight variable for entity {} nor alternative weights. Switch to unweighted'.format(entity_key))

    # Compute the weigthed number of zeros or non zeros
    value_by_simulation = dict(after = after, before = before)
    stats_by_simulation = dict()
    for simulation_prefix, value in value_by_simulation.items():
        stats = dict()
        stats["count_zero"] = (
            weight.astype("float64")
            * (
                (absolute_minimal_detected_variation > np.abs(value))
                )
            ).sum()
        stats["count_non_zero"] = sum(weight.astype("float64")) - stats["count_zero"]
        stats_by_simulation[simulation_prefix] = stats
        del stats

    # Compute the number of entity above or below after
    after_value = after
    before_value = before
    with np.errstate(divide="ignore", invalid="ignore"):
        above_after = ((after_value - before_value) / np.abs(before_value)) > relative_minimal_detected_variation
    almost_zero_before = np.abs(before_value) < absolute_minimal_detected_variation
    above_after[almost_zero_before * (after_value >= 0)] = (
        after_value >= absolute_minimal_detected_variation
        )[almost_zero_before * (after_value >= 0)]
    with np.errstate(divide="ignore", invalid="ignore"):
        below_after = ((after_value - before_value) / np.abs(before_value)) < -relative_minimal_detected_variation
    below_after[almost_zero_before * (after_value < 0)] = (
        after_value < -absolute_minimal_detected_variation
        )[almost_zero_before * (after_value < 0)]

    # Check if there is a secret violation, without weights
    if observations_threshold is not None:
        not_legit_below = (below_after.sum() < observations_threshold) & (below_after.sum() > 0)
        not_legit_above = (above_after.sum() < observations_threshold) & (above_after.sum() > 0)
        if not_legit_below | not_legit_above:
            raise SecretViolationError("Not enough observations involved")

    # Apply weights
    above_after_count = (above_after.astype("float64") * weight.astype("float64")).sum()
    below_after_count = (below_after.astype("float64") * weight.astype("float64")).sum()
    total = sum(weight)
    neutral = total - above_after_count - below_after_count

    return {
        "total": total,
        "non_zero_before": stats_by_simulation["before"]["count_non_zero"],
        "non_zero_after": stats_by_simulation["after"]["count_non_zero"],
        "above_after": above_after_count,
        "lower_after": below_after_count,
        "neutral": neutral,
        "tolerance_factor_used": relative_minimal_detected_variation,
        "weight_factor": 1,
        }


def init_entity_data(simulation: Simulation, entity: Entity, filtered_input_data_frame: pd.DataFrame, period: Period,
        used_as_input_variables_by_entity: Dict):
    """
    Initialize entity in simulation at some period with input provided by a dataframe.

    Args:
        simulation (Simulation): The simulation to initialize.
        entity (Entity): The entity which variables to initialize.
        filtered_input_data_frame (pd.DataFrame): The dataframe with the variables values.
        period (Period): The period to initialize.
        used_as_input_variables_by_entity (Dict): The variable to be used to initialize each entity.
    """
    used_as_input_variables = used_as_input_variables_by_entity[entity.key]
    input_data_frame = filtered_input_data_frame
    # input_data_frame = self.filter_input_variables(input_data_frame = input_data_frame)
    diagnose_variable_mismatch(used_as_input_variables, input_data_frame)

    for column_name, column_serie in input_data_frame.items():
        variable_instance = simulation.tax_benefit_system.variables.get(column_name)
        if variable_instance is None:
            log.info(f"Ignoring {column_name} in input data")
            continue

        if variable_instance.entity.key != entity.key:
            log.info("Ignoring variable {} which is not part of entity {} but {}".format(
                column_name, entity.key, variable_instance.entity.key))
            continue
        init_variable_in_entity(simulation, entity.key, column_name, column_serie, period)


def inflate(simulation: Simulation, inflator_by_variable: Optional[Dict] = None, period: Optional[Union[int, str, Period]] = None,
        target_by_variable: Optional[Dict] = None):
    tax_benefit_system = simulation.tax_benefit_system
    for variable_name in set(inflator_by_variable.keys()).union(set(target_by_variable.keys())):
        assert variable_name in tax_benefit_system.variables, \
            "Variable {} is not a valid variable of the tax-benefit system".format(variable_name)
        if variable_name in target_by_variable:
            inflator = inflator_by_variable[variable_name] = \
                target_by_variable[variable_name] / simulation.compute_aggregate(
                    variable = variable_name, period = period)
            log.info('Using {} as inflator for {} to reach the target {} '.format(
                inflator, variable_name, target_by_variable[variable_name]))
        else:
            assert variable_name in inflator_by_variable, 'variable_name is not in inflator_by_variable'
            log.info('Using inflator {} for {}.  The target is thus {}'.format(
                inflator_by_variable[variable_name],
                variable_name, inflator_by_variable[variable_name] * simulation.compute_aggregate(
                    variable = variable_name, period = period)
                ))
            inflator = inflator_by_variable[variable_name]

        array = simulation.calculate_add(variable_name, period = period)
        assert array is not None
        simulation.delete_arrays(variable_name, period = period)  # delete existing arrays
        simulation.set_input(variable_name, period, inflator * array)  # insert inflated array


def init_simulation(tax_benefit_system, period, data):
    builder = SimulationBuilder()
    builder.create_entities(tax_benefit_system)

    collection = data.get("collection")
    custom_input_data_frame = data.get("custom_input_data_frame", do_nothing)
    data_year = data.get("data_year")
    survey = data.get('survey')
    config_files_directory = data.get("config_files_directory")
    builder.used_as_input_variables = data.get("used_as_input_variables")
    builder.id_variable_by_entity_key = data.get("id_variable_by_entity_key")
    builder.role_variable_by_entity_key = data.get("role_variable_by_entity_key")
    builder.tax_benefit_system = tax_benefit_system

    default_source_types = [
        'input_data_frame',
        'input_data_table',
        'input_data_frame_by_entity',
        'input_data_frame_by_entity_by_period',
        'input_data_table_by_entity_by_period',
        'input_data_table_by_period',
        ]
    source_types = [
        source_type_
        for source_type_ in default_source_types
        if data.get(source_type_, None) is not None
        ]
    assert len(source_types) < 2, "There are too many data source types"
    assert len(source_types) >= 1, "There should be one data source type included in {}".format(
        default_source_types)
    source_type = source_types[0]
    source = data[source_type]

    if source_type == 'input_data_frame_by_entity':
        assert data_year is not None
        source_type = 'input_data_frame_by_entity_by_period'
        source = {periods.period(data_year): source}

    input_data_survey_prefix = data.get("input_data_survey_prefix") if data is not None else None

    if source_type == 'input_data_frame':
        simulation = builder.init_all_entities(source, period)

    if source_type == 'input_data_table':
        # Case 1: fill simulation with a unique input_data_frame given by the attribute
        if input_data_survey_prefix is not None:
            openfisca_survey_collection = SurveyCollection.load(collection = collection)
            openfisca_survey = openfisca_survey_collection.get_survey("{}_{}".format(
                input_data_survey_prefix, data_year))
            input_data_frame = openfisca_survey.get_values(table = "input").reset_index(drop = True)
        else:
            NotImplementedError

        custom_input_data_frame(input_data_frame, period = period)
        simulation = builder.init_all_entities(input_data_frame, builder, period)  # monolithic dataframes

    elif source_type == 'input_data_table_by_period':
        # Case 2: fill simulation with input_data_frame by period containing all entity variables
        input_data_table_by_period = data.get("input_data_table_by_period")
        for period, table in input_data_table_by_period.items():
            period = periods.period(period)
            log.debug('From survey {} loading table {}'.format(survey, table))
            input_data_frame = load_table(config_files_directory = config_files_directory, collection = collection, survey = survey, input_data_survey_prefix = input_data_survey_prefix, table = table)
            custom_input_data_frame(input_data_frame, period = period)
            simulation = builder.init_all_entities(input_data_frame, builder, period)  # monolithic dataframes

    elif source_type == 'input_data_frame_by_entity_by_period':
        for period, input_data_frame_by_entity in source.items():
            period = periods.period(period)
            for entity in tax_benefit_system.entities:
                input_data_frame = input_data_frame_by_entity.get(entity.key)
                if input_data_frame is None:
                    continue
                custom_input_data_frame(input_data_frame, period = period, entity = entity.key)
                builder.init_entity_structure(entity, input_data_frame)  # TODO complete args

        simulation = builder.build(tax_benefit_system)
        simulation.id_variable_by_entity_key = builder.id_variable_by_entity_key  # Should be propagated to enhanced build

        for period, input_data_frame_by_entity in source.items():
            for entity in tax_benefit_system.entities:
                input_data_frame = input_data_frame_by_entity.get(entity.key)
                if input_data_frame is None:
                    log.debug("No input_data_frame found for entity {} at period {}".format(entity, period))
                    continue
                custom_input_data_frame(input_data_frame, period = period, entity = entity.key)
                simulation.init_entity_data(entity, input_data_frame, period, builder.used_as_input_variables_by_entity)

    elif source_type == 'input_data_table_by_entity_by_period':
        # Case 3: fill simulation with input_data_table by entity_by_period containing a dictionnary
        # of all periods containing a dictionnary of entity variables
        input_data_table_by_entity_by_period = source
        simulation = None
        for period, input_data_table_by_entity in input_data_table_by_entity_by_period.items():
            period = periods.period(period)

            if simulation is None:
                for entity in tax_benefit_system.entities:
                    table = input_data_table_by_entity.get(entity.key)
                    if table is None:
                        continue
                    if survey is not None:
                        input_data_frame = load_table(config_files_directory = config_files_directory, collection = collection, survey = survey, table = table)
                    else:
                        input_data_frame = load_table(config_files_directory = config_files_directory, collection = collection, survey = 'input', table = table)
                    custom_input_data_frame(input_data_frame, period = period, entity = entity.key)
                    builder.init_entity_structure(entity, input_data_frame)  # TODO complete args

                simulation = builder.build(tax_benefit_system)
                simulation.id_variable_by_entity_key = builder.id_variable_by_entity_key  # Should be propagated to enhanced build

            for entity in tax_benefit_system.entities:
                table = input_data_table_by_entity.get(entity.key)
                if table is None:
                    continue
                if survey is not None:
                    input_data_frame = load_table(config_files_directory = config_files_directory, collection = collection, survey = survey, table = table)
                else:
                    input_data_frame = load_table(config_files_directory = config_files_directory, collection = collection, survey = 'input', table = table)
                custom_input_data_frame(input_data_frame, period = period, entity = entity.key)
                simulation.init_entity_data(entity, input_data_frame, period, builder.used_as_input_variables_by_entity)
    else:
        pass

    if data_year is not None:
        simulation.period = periods.period(data_year)

    return simulation


def init_variable_in_entity(simulation: Simulation, entity, variable_name, series, period):
    variable = simulation.tax_benefit_system.variables[variable_name]

    # np.issubdtype cannot handles categorical variables
    if (not pd.api.types.is_categorical_dtype(series)) and np.issubdtype(series.values.dtype, np.floating):
        if series.isnull().any():
            log.debug('There are {} NaN values for {} non NaN values in variable {}'.format(
                series.isnull().sum(), series.notnull().sum(), variable_name))
            log.debug('We convert these NaN values of variable {} to {} its default value'.format(
                variable_name, variable.default_value))
            series.fillna(variable.default_value, inplace = True)
        assert series.notnull().all(), \
            'There are {} NaN values for {} non NaN values in variable {}'.format(
                series.isnull().sum(), series.notnull().sum(), variable_name)

    enum_variable_imputed_as_enum = (
        variable.value_type == Enum
        and (
            pd.api.types.is_categorical_dtype(series)
            or not (
                np.issubdtype(series.values.dtype, np.integer)
                or np.issubdtype(series.values.dtype, float)
                )
            )
        )

    if enum_variable_imputed_as_enum:
        if series.isnull().any():
            log.debug('There are {} NaN values ({}% of the array) in variable {}'.format(
                series.isnull().sum(), series.isnull().mean() * 100, variable_name))
            log.debug('We convert these NaN values of variable {} to {} its default value'.format(
                variable_name, variable.default_value._name_))
            series.fillna(variable.default_value._name_, inplace = True)
        possible_values = variable.possible_values
        index_by_category = dict(zip(
            possible_values._member_names_,
            range(len(possible_values._member_names_))
            ))
        series.replace(index_by_category, inplace = True)

    if series.values.dtype != variable.dtype:
        log.debug(
            'Converting {} from dtype {} to {}'.format(
                variable_name, series.values.dtype, variable.dtype)
            )

    array = series.values.astype(variable.dtype)
    # TODO is the next line needed ?
    # Might be due to values returning also ndarray like objects
    # for instance for categories or
    np_array = np.array(array, dtype = variable.dtype)
    if variable.definition_period == YEAR and period.unit == MONTH:
        # Some variables defined for a year are present in month/quarter dataframes
        # Cleaning the dataframe would probably be better in the long run
        log.warn(f"Trying to set a monthly value for variable {variable_name}, which is defined on a year. The  montly values you provided will be summed.")

        if simulation.get_array(variable_name, period.this_year) is not None:
            array_sum = simulation.get_array(variable_name, period.this_year) + np_array
            simulation.set_input(variable_name, period.this_year, array_sum)
        else:
            simulation.set_input(variable_name, period.this_year, np_array)

    else:
        simulation.set_input(variable_name, period, np_array)


def new_from_tax_benefit_system(
        tax_benefit_system: TaxBenefitSystem,
        debug: bool = False,
        trace: bool = False,
        data: Dict = None,
        memory_config: MemoryConfig = None,
        period: Optional[Union[int, str, Period]] = None,
        custom_initialize: Callable = None,
        ) -> Simulation:
    """
    Create and initialize a simulation from a tax and benefit system and data.

    Args:
        tax_benefit_system (TaxBenefitSystem): The tax and benefit system
        debug (bool, optional): Whether to activate debugging. Defaults to False.
        trace (bool, optional): Whether to activate tracing. Defaults to False.
        data (Dict, optional): The information about data. Defaults to None.
        memory_config (MemoryConfig, optional): The memory handling config. Defaults to None.
        period (Optional[Union[int, str, Period]], optional): The period of the simulation. Defaults to None.
        custom_initialize (Callable, optional): The post-processing initialization function. Defaults to None.

    Returns:
        Simulation: The completely initialized function
    """

    simulation = Simulation.init_simulation(tax_benefit_system, period, data)
    simulation.debug = debug
    simulation.trace = trace
    simulation.opt_out_cache = True if simulation.tax_benefit_system.cache_blacklist is not None else False
    simulation.memory_config = memory_config

    if custom_initialize:
        custom_initialize(simulation)

    return simulation


def print_memory_usage(simulation: Simulation):
    """
    Print memory usage.

    Args:
        simulation (Simulation): The simulation which memory usage is to be printed
    """
    memory_usage_by_variable = simulation.get_memory_usage()['by_variable']
    try:
        usage_stats = simulation.tracer.usage_stats
    except AttributeError:
        log.warning("The simulation trace mode is not activated. You need to activate it to get stats about variable usage (hits).")
        usage_stats = None
    infos_lines = list()

    for variable, infos in memory_usage_by_variable.items():
        hits = usage_stats[variable]['nb_requests'] if usage_stats else None
        infos_lines.append((
            infos['total_nb_bytes'],
            variable, "{}: {} periods * {} cells * item size {} ({}) = {} with {} hits".format(
                variable,
                infos['nb_arrays'],
                infos['nb_cells_by_array'],
                infos['cell_size'],
                infos['dtype'],
                humanize.naturalsize(infos['total_nb_bytes'], gnu = True),
                hits,
                )
            ))
    infos_lines.sort()
    for _, _, line in infos_lines:
        print(line.rjust(100))  # noqa analysis:ignore


def set_weight_variable_by_entity(
        simulation: Simulation,
        weight_variable_by_entity: Dict,
        ):
    """
    Set weight variable for each entity.

    Args:
        simulation (Simulation): The simulation concerned.
        weight_variable_by_entity (Dict): The weight variable for each entity.
    """
    simulation.weight_variable_by_entity = weight_variable_by_entity


def summarize_variable(simulation: Simulation, variable = None, weighted = False, force_compute = False):
    """Print a summary of a variable including its memory usage.

    Args:
        variable(string): The variable being summarized
        use_baseline(bool): The tax-benefit-system considered
        weighted(bool): Whether the produced statistics should be weigthted or not
        force_compute(bool): Whether the computation of the variable should be forced

    # Example:
    #     >>> from openfisca_survey_manager.tests.test_scenario import create_randomly_initialized_survey_scenario
    #     >>> survey_scenario = create_randomly_initialized_survey_scenario()
    #     >>> survey_scenario.summarize_variable(variable = "housing_occupancy_status", force_compute = True)
    #     <BLANKLINE>
    #     housing_occupancy_status: 1 periods * 5 cells * item size 2 (int16, default = HousingOccupancyStatus.tenant) = 10B
    #     Details:
    #     2017-01: owner = 0.00e+00 (0.0%), tenant = 5.00e+00 (100.0%), free_lodger = 0.00e+00 (0.0%), homeless = 0.00e+00 (0.0%).
    #     >>> survey_scenario.summarize_variable(variable = "rent", force_compute = True)
    #     <BLANKLINE>
    #     rent: 2 periods * 5 cells * item size 4 (float32, default = 0) = 40B
    #     Details:
    #     2017-01: mean = 562.385107421875, min = 156.01864624023438, max = 950.7142944335938, mass = 2.81e+03, default = 0.0%, median = 598.6585083007812
    #     2018-01: mean = 562.385107421875, min = 156.01864624023438, max = 950.7142944335938, mass = 2.81e+03, default = 0.0%, median = 598.6585083007812
    #     >>> survey_scenario.tax_benefit_system.neutralize_variable('age')
    #     >>> survey_scenario.summarize_variable(variable = "age")
    #     <BLANKLINE>
    #     age: neutralized variable (int64, default = 0)
    """
    tax_benefit_system = simulation.tax_benefit_system
    variable_instance = tax_benefit_system.variables.get(variable)
    assert variable_instance is not None, "{} is not a valid variable".format(variable)

    default_value = variable_instance.default_value
    value_type = variable_instance.value_type

    if variable_instance.is_neutralized:
        print("")  # noqa analysis:ignore
        print("{}: neutralized variable ({}, default = {})".format(variable, str(np.dtype(value_type)), default_value))  # noqa analysis:ignore
        return

    if weighted:
        weight_variable = simulation.weight_variable_by_entity[variable_instance.entity.key]
        weights = simulation.calculate(weight_variable, simulation.period)

    infos = simulation.get_memory_usage(variables = [variable])['by_variable'].get(variable)
    if not infos:
        if force_compute:
            simulation.adaptative_calculate_variable(variable = variable, period = simulation.period)
            simulation.summarize_variable(variable = variable, weighted = weighted)
            return
        else:
            print("{} is not computed yet. Use keyword argument force_compute = True".format(variable))  # noqa analysis:ignore
            return

    header_line = "{}: {} periods * {} cells * item size {} ({}, default = {}) = {}".format(
        variable,
        infos['nb_arrays'],
        infos['nb_cells_by_array'],
        infos['cell_size'],
        str(np.dtype(infos['dtype'])),
        default_value,
        humanize.naturalsize(infos['total_nb_bytes'], gnu = True),
        )
    print("")  # noqa analysis:ignore
    print(header_line)  # noqa analysis:ignore
    print("Details:")  # noqa analysis:ignore
    holder = simulation.get_holder(variable)
    if holder is not None:
        if holder.variable.definition_period == ETERNITY:
            array = holder.get_array(ETERNITY)
            print("permanent: mean = {}, min = {}, max = {}, median = {}, default = {:.1%}".format(  # noqa analysis:ignore
                # Need to use float to avoid hit the int16/int32 limit. np.average handles it without conversion
                array.astype(float).mean() if not weighted else np.average(array, weights = weights),
                array.min(),
                array.max(),
                np.median(array.astype(float)),
                (
                    (array == default_value).sum() / len(array)
                    if not weighted
                    else ((array == default_value) * weights).sum() / weights.sum()
                    )
                ))
        else:
            for period in sorted(simulation.get_known_periods(variable)):
                array = holder.get_array(period)
                if array.shape == ():
                    print("{}: always = {}".format(period, array))  # noqa analysis:ignore
                    continue

                if value_type == Enum:
                    possible_values = variable_instance.possible_values
                    categories_by_index = dict(zip(
                        range(len(possible_values._member_names_)),
                        possible_values._member_names_
                        ))
                    categories_type = pd.api.types.CategoricalDtype(categories = possible_values._member_names_, ordered = True)
                    df = pd.DataFrame({variable: array}).replace(categories_by_index).astype(categories_type)
                    df['weights'] = weights if weighted else 1
                    groupby = df.groupby(variable)['weights'].sum()
                    total = groupby.sum()
                    expr = [" {} = {:.2e} ({:.1%})".format(index, row, row / total) for index, row in groupby.items()]
                    print("{}:{}.".format(period, ",".join(expr)))  # noqa analysis:ignore
                    continue

                print("{}: mean = {}, min = {}, max = {}, mass = {:.2e}, default = {:.1%}, median = {}".format(  # noqa analysis:ignore
                    period,
                    array.astype(float).mean() if not weighted else np.average(array, weights = weights),
                    array.min(),
                    array.max(),
                    array.astype(float).sum() if not weighted else np.sum(array * weights),
                    (
                        (array == default_value).sum() / len(array)
                        if not weighted
                        else ((array == default_value) * weights).sum() / weights.sum()
                        ),
                    np.median(array),
                    ))


# Monkey patching

Simulation.adaptative_calculate_variable = adaptative_calculate_variable
Simulation.compute_aggregate = compute_aggregate
Simulation.compute_pivot_table = compute_pivot_table
Simulation.create_data_frame_by_entity = create_data_frame_by_entity
Simulation.compute_quantiles = compute_quantiles
Simulation.compute_winners_loosers = compute_winners_loosers
Simulation.new_from_tax_benefit_system = new_from_tax_benefit_system
Simulation.inflate = inflate
Simulation.init_entity_data = init_entity_data
Simulation.init_simulation = init_simulation
Simulation.init_variable_in_entity = init_variable_in_entity
Simulation.print_memory_usage = print_memory_usage
Simulation.set_weight_variable_by_entity = set_weight_variable_by_entity
Simulation.summarize_variable = summarize_variable



import os


from openfisca_survey_manager import openfisca_survey_manager_location
from openfisca_survey_manager.survey_collections import SurveyCollection
from openfisca_survey_manager.surveys import Survey


def test_survey():
    name = 'fake'
    data_dir = os.path.join(
        openfisca_survey_manager_location,
        'openfisca_survey_manager',
        'tests',
        'data_files',
        )

    survey_collection = SurveyCollection(
        name = name,
        config_files_directory = data_dir,
        json_file_path = os.path.join(data_dir, 'fake.json')
        )

    saved_fake_survey_hdf5_file_path = os.path.join(data_dir, 'fake.hdf5')
    saved_fake_survey_file_path = os.path.join(data_dir, 'help.sas7bdat')
    survey = Survey(
        hdf5_file_path = saved_fake_survey_hdf5_file_path,
        name = 'fake_survey',
        sas_files = [saved_fake_survey_file_path],
        survey_collection = survey_collection,
        )
    survey.insert_table(name = 'help')
    survey.fill_hdf(source_format = 'sas')


def test_survey_load():
    survey_name = 'test_set_table_in_survey_2021'
    collection = 'fake'
    data_dir = os.path.join(
        openfisca_survey_manager_location,
        'openfisca_survey_manager',
        'tests',
        'data_files',
        )
    survey_collection = SurveyCollection.load(
        collection=collection, config_files_directory=data_dir
        )
    survey = survey_collection.get_survey(survey_name)
    for table_name, _ in survey.tables.items():
        assert table_name == "foyer_2021"


if __name__ == '__main__':
    test_survey()



"""Abstract survey scenario definition."""

import logging
import os
import numpy as np
import pandas as pd
from typing import Dict, List, Optional, Union


from openfisca_core import periods
from openfisca_core.types import Array, Period, TaxBenefitSystem
from openfisca_survey_manager.simulations import Simulation  # noqa analysis:ignore
from openfisca_core.periods import MONTH, YEAR
from openfisca_core.tools.simulation_dumper import dump_simulation, restore_simulation


from openfisca_survey_manager.calibration import Calibration
from openfisca_survey_manager.surveys import Survey

log = logging.getLogger(__name__)


class AbstractSurveyScenario(object):
    """Abstract survey scenario."""

    cache_blacklist = None
    collection = None
    debug = False
    filtering_variable_by_entity = None
    id_variable_by_entity_key = None
    inflator_by_variable = None  # factor used to inflate variable total
    input_data_frame = None
    input_data_table_by_entity_by_period = None
    input_data_table_by_period = None
    non_neutralizable_variables = None
    period = None
    role_variable_by_entity_key = None
    simulations = None
    target_by_variable = None  # variable total target to inflate to
    tax_benefit_systems = None
    trace = False
    used_as_input_variables = None
    used_as_input_variables_by_entity = None
    variation_factor = .03  # factor used to compute variation when estimating marginal tax rate
    varying_variable = None
    weight_variable_by_entity = None

    def build_input_data(self, **kwargs):
        """Build input data."""
        NotImplementedError

    def calculate_series(self, variable, period = None, simulation = None):
        """Compute variable values for period for a given simulation.

        Args:
          variable(str, optional): Variable to compute
          period(Period, optional): Period, defaults to None
          simulation(str, optional): Simulation to use

        Returns:
          pandas.Series: Variable values

        """
        return pd.Series(
            data = self.calculate_variable(variable, period, simulation = simulation),
            name = variable,
            )

    def calculate_variable(self, variable, period = None, simulation = None):
        """Compute variable values for period for a given simulation.

        Args:
          variable(str, optional): Variable to compute
          period(Period, optional): Period, defaults to None
          simulation(str, optional): Simulation to use

        Returns:
          numpy.ndarray: Variable values

        """
        if simulation is None:
            assert len(self.simulations.keys()) == 1
            simulation = list(self.simulations.values())[0]
        else:
            simulation = self.simulations[simulation]
        assert simulation is not None
        return simulation.adaptative_calculate_variable(variable, period = period)

    def calibrate(self, period: int = None, target_margins_by_variable: dict = None, parameters: dict = None,
            target_entity_count: float = None, entity: str = None):
        """Calibrate the scenario data.

        Args:
            period (int, optionnal): Period of calibration. Defaults to scenario.year
            target_margins_by_variable (dict, optional): Variable targets margins. Defaults to None.
            parameters (dict, optional): Calibration parameters. Defaults to None.
            target_entity_count (float, optional): Total population target. Defaults to None.
            entity (str): Entity specified when no variable comes with a target margins but `target_entity_count` is not None.
        """
        survey_scenario = self

        if period is None:
            assert survey_scenario.period is not None
            period = survey_scenario.period

        if parameters is not None:
            assert parameters['method'] in ['linear', 'raking ratio', 'logit'], \
                "Incorect parameter value: method should be 'linear', 'raking ratio' or 'logit'"
            if parameters['method'] == 'logit':
                assert parameters['invlo'] is not None
                assert parameters['up'] is not None
        else:
            parameters = dict(method = 'logit', up = 3, invlo = 3)

        # TODO: filtering using filtering_variable_by_entity
        for simulation in self.simulations.values():
            if simulation is None:
                continue
            calibration = Calibration(
                simulation,
                target_margins_by_variable,
                period,
                target_entity_count = target_entity_count,
                entity = entity,
                parameters = parameters,
                # filter_by = self.filter_by,
                )
            calibration.calibrate(inplace = True)
            simulation.calibration = calibration

    def compute_aggregate(self, variable: str = None, aggfunc: str = 'sum', filter_by: str = None,
            period: Optional[Union[int, str, Period]] = None,
            simulation: str = None, baseline_simulation: str = None, missing_variable_default_value = np.nan,
            weighted: bool = True, alternative_weights: Optional[Union[str, int, float, Array]] = None):
        """Compute variable aggregate.

        Args:
            variable (str, optional): Variable to aggregate. Defaults to None.
            aggfunc (str, optional): Aggregation function. Defaults to 'sum'.
            filter_by (str, optional): Filter variable or expression to use. Defaults to None.
            period (Optional[Union[int, str, Period]], optional): Period. Defaults to None.
            simulation(str, optional): Simulation to use
            baseline_simulation(str, optional): Baseline simulation to use when computing a difference
            missing_variable_default_value (optional): Value to use for missing values. Defaults to np.nan.
            weighted (bool, optional): Whether to weight the variable or not. Defaults to True.
            alternative_weights (Optional[Union[str, int, float, Array]], optional): Alternative weigh to use. Defaults to None.
            filtering_variable_by_entity (Dict, optional): Filtering variable by entity. Defaults to None.

        Returns:
            float: Aggregate
        """
        assert aggfunc in ['count', 'mean', 'sum', 'count_non_zero']
        assert period is not None
        assert variable is not None
        if simulation is None:
            assert len(self.simulations.keys()) == 1
            simulation = list(self.simulations.values())[0]
        else:
            simulation = self.simulations[simulation]

        assert simulation is not None, f"Missing {simulation} simulation"

        if baseline_simulation:
            baseline_simulation = self.simulations[baseline_simulation]
            return (
                simulation.compute_aggregate(
                    variable = variable,
                    aggfunc = aggfunc,
                    filter_by = filter_by,
                    period = period,
                    missing_variable_default_value = missing_variable_default_value,
                    weighted = weighted,
                    alternative_weights = alternative_weights,
                    filtering_variable_by_entity = self.filtering_variable_by_entity,
                    )
                - baseline_simulation.compute_aggregate(
                    variable = variable,
                    aggfunc = aggfunc,
                    filter_by = filter_by,
                    period = period,
                    missing_variable_default_value = missing_variable_default_value,
                    weighted = weighted,
                    alternative_weights = alternative_weights,
                    filtering_variable_by_entity = self.filtering_variable_by_entity,
                    )
                )

        return simulation.compute_aggregate(
            variable = variable,
            aggfunc = aggfunc,
            filter_by = filter_by,
            period = period,
            missing_variable_default_value = missing_variable_default_value,
            weighted = weighted,
            alternative_weights = alternative_weights,
            filtering_variable_by_entity = self.filtering_variable_by_entity,
            )

    def compute_quantiles(self, simulation: Simulation, variable: str, nquantiles: int = None,
            period: Optional[Union[int, str, Period]] = None, filter_by = None, weighted: bool = True,
            alternative_weights = None, filtering_variable_by_entity = None) -> List[float]:
        """
        Compute quantiles of a variable.

        Args:
            simulation (Simulation, optional): Simulation to be used. Defaults to None.
            variable (str, optional): Variable which quantiles are computed. Defaults to None.
            nquantiles (int, optional): Number of quantiles. Defaults to None.
            period (Optional[Union[int, str, Period]], optional): Period. Defaults to None.
            missing_variable_default_value (optional): Value to use for missing values. Defaults to np.nan.
            weighted (bool, optional): Whether to weight the variable or not. Defaults to True.
            alternative_weights (Optional[Union[str, int, float, Array]], optional): Alternative weigh to use. Defaults to None.
            filtering_variable_by_entity (Dict, optional): Filtering variable by entity. Defaults to None.

        Returns:
        List(float): The quantiles values
        """
        assert variable is not None
        assert nquantiles is not None
        simulation = self.simulations[simulation]
        assert simulation is not None, f"Missing {simulation} simulation"

        return simulation.compute_quantiles(
            variable = variable,
            period = period,
            nquantiles = nquantiles,
            filter_by = filter_by,
            weighted = weighted,
            alternative_weights = alternative_weights,
            )

    def compute_marginal_tax_rate(self, target_variable: str, period: Optional[Union[int, str, Period]], simulation: str = None,
            value_for_zero_varying_variable: float = 0.0) -> Array:
        """
        Compute marginal a rate of a target (MTR) with respect to a varying variable.

        Args:
            target_variable (str): the variable which marginal tax rate is computed
            period (Optional[Union[int, str, Period]], optional): Period. Defaults to None.
            simulation(str, optional): Simulation to use
            value_for_zero_varying_variable (float, optional): value of MTR when the varying variable is zero. Defaults to 0.

        Returns:
            numpy.array: Vector of marginal rates
        """
        varying_variable = self.varying_variable
        if simulation is None:
            assert len(self.simulations.keys()) == 2
            simulation_name = [name for name in self.simulations.keys() if not name.startswith("_modified_")][0]
            simulation = self.simulations[simulation_name]
        else:
            simulation_name = simulation
            simulation = self.simulations[simulation_name]

        modified_simulation = self.simulations[f"_modified_{simulation_name}"]

        variables = simulation.tax_benefit_system.variables
        assert target_variable in variables

        variables_belong_to_same_entity = (
            variables[varying_variable].entity.key == variables[target_variable].entity.key
            )
        varying_variable_belongs_to_person_entity = variables[varying_variable].entity.is_person

        assert variables_belong_to_same_entity or varying_variable_belongs_to_person_entity

        if variables_belong_to_same_entity:
            modified_varying = modified_simulation.calculate_add(varying_variable, period = period)
            varying = simulation.calculate_add(varying_variable, period = period)
        else:
            target_variable_entity_key = variables[target_variable].entity.key

            def cast_to_target_entity(simulation: Simulation):
                population = simulation.populations[target_variable_entity_key]
                df = (pd.DataFrame(
                    {
                        'members_entity_id': population._members_entity_id,
                        varying_variable: simulation.calculate_add(varying_variable, period = period)
                        }
                    ).groupby('members_entity_id').sum())
                varying_variable_for_target_entity = df.loc[population.ids, varying_variable].values
                return varying_variable_for_target_entity

            modified_varying = cast_to_target_entity(modified_simulation)
            varying = cast_to_target_entity(simulation)

        modified_target = modified_simulation.calculate_add(target_variable, period = period)
        target = simulation.calculate_add(target_variable, period = period)

        numerator = modified_target - target
        denominator = modified_varying - varying
        marginal_rate = 1 - np.divide(
            numerator,
            denominator,
            out = np.full_like(numerator, value_for_zero_varying_variable, dtype = np.floating),
            where = (denominator != 0)
            )

        return marginal_rate

    def compute_pivot_table(self, aggfunc = 'mean', columns = None, baseline_simulation = None, filter_by = None, index = None,
            period = None, simulation = None, difference = False, use_baseline_for_columns = None, values = None,
            missing_variable_default_value = np.nan, concat_axis = None, weighted = True, alternative_weights = None):
        """Compute a pivot table of agregated values casted along specified index and columns.

        Args:
            aggfunc(str, optional): Aggregation function, defaults to 'mean'
            columns(list, optional): Variable(s) in columns, defaults to None
            difference(bool, optional): Compute difference, defaults to False
            filter_by(str, optional): Boolean variable to filter by, defaults to None
            index(list, optional): Variable(s) in index (lines), defaults to None
            period(Period, optional): Period, defaults to None
            simulation(str, optional): Simulation to use
            baseline_simulation(str, optional): Baseline simulation to use when computing a difference
            use_baseline_for_columns(bool, optional): Use columns from baseline columns values, defaults to None
            values(list, optional): Aggregated variable(s) within cells, defaults to None
            missing_variable_default_value(float, optional): Default value for missing variables, defaults to np.nan
            concat_axis(int, optional): Axis to concatenate along (index = 0, columns = 1), defaults to None
            weighted(bool, optional): Whether to weight te aggregates (Default value = True)
            alternative_weights(str or int or float, optional): Weight variable name or numerical value. Use Simulation's weight_variable_by_entity if None, and if the later is None uses 1 ((Default value = None)

        Returns:
            pd.DataFrame: Pivot table

        """
        assert (not difference) or (baseline_simulation is not None), "Can't have difference when not baseline simulation"

        simulation = self.simulations[simulation]
        if baseline_simulation:
            baseline_simulation = self.simulations[baseline_simulation]

        filtering_variable_by_entity = self.filtering_variable_by_entity

        return simulation.compute_pivot_table(
            baseline_simulation = baseline_simulation,
            aggfunc = aggfunc,
            columns = columns,
            difference = difference,
            filter_by = filter_by,
            index = index,
            period = period,
            use_baseline_for_columns = use_baseline_for_columns,
            values = values,
            missing_variable_default_value = missing_variable_default_value,
            concat_axis = concat_axis,
            weighted = weighted,
            alternative_weights = alternative_weights,
            filtering_variable_by_entity = filtering_variable_by_entity,
            )

    def compute_winners_loosers(self, variable,
            simulation,
            baseline_simulation = None,
            filter_by = None,
            period = None,
            absolute_minimal_detected_variation = 0,
            relative_minimal_detected_variation = .01,
            observations_threshold = None,
            weighted = True,
            alternative_weights = None):

        simulation = self.simulations[simulation]
        if baseline_simulation:
            baseline_simulation = self.simulations[baseline_simulation]

        return simulation.compute_winners_loosers(
            baseline_simulation,
            variable = variable,
            filter_by = filter_by,
            period = period,
            absolute_minimal_detected_variation = absolute_minimal_detected_variation,
            relative_minimal_detected_variation = relative_minimal_detected_variation,
            observations_threshold = observations_threshold,
            weighted = weighted,
            alternative_weights = alternative_weights,
            filtering_variable_by_entity = self.filtering_variable_by_entity,
            )

    def create_data_frame_by_entity(self, variables = None, expressions = None, filter_by = None, index = False,
            period = None, simulation = None, merge = False):
        """Create dataframe(s) of computed variable for every entity (eventually merged in a unique dataframe).

        Args:
          variables(list, optional): Variable to compute, defaults to None
          expressions(str, optional): Expressions to compute, defaults to None
          filter_by(str, optional): Boolean variable or expression, defaults to None
          index(bool, optional): Index by entity id, defaults to False
          period(Period, optional): Period, defaults to None
          simulation(str, optional): Simulation to use
          merge(bool, optional): Merge all the entities in one data frame, defaults to False

        Returns:
          dict or pandas.DataFrame: Dictionnary of dataframes by entities or dataframe with all the computed variables

        """
        if simulation is None:
            assert len(self.simulations.keys()) == 1
            simulation = list(self.simulations.values())[0]
        else:
            simulation = self.simulations[simulation]

        return simulation.create_data_frame_by_entity(
            variables = variables,
            expressions = expressions,
            filter_by = filter_by,
            index = index,
            period = period,
            merge = merge,
            )

    def custom_input_data_frame(self, input_data_frame, **kwargs):
        """Customize input data frame.

        Args:
          input_data_frame: original input data frame.
          kwargs: keyword arguments.
        """
        pass

    def dump_data_frame_by_entity(self, variables = None, survey_collection = None, survey_name = None):
        assert survey_collection is not None
        assert survey_name is not None
        assert variables is not None
        openfisca_data_frame_by_entity = self.create_data_frame_by_entity(variables = variables)
        for entity_key, data_frame in openfisca_data_frame_by_entity.items():
            survey = Survey(name = survey_name)
            survey.insert_table(name = entity_key, data_frame = data_frame)
            survey_collection.surveys.append(survey)
            survey_collection.dump(collection = "openfisca")

    def dump_simulations(self, directory: str):
        """
        Dump simulations.

        Args:
            directory (str, optional): Dump directory
        """
        assert directory is not None
        use_sub_directories = True if len(self.simulations) >= 2 else False

        if use_sub_directories:
            for simulation_name, simulation in self.simulations.items():
                dump_simulation(simulation, directory = os.path.join(directory, simulation_name))
        else:
            assert len(self.simulations.keys()) == 1
            simulation = list(self.simulations.values())[0]
            dump_simulation(simulation, directory)

    def generate_performance_data(self, output_dir: str):
        if not self.trace:
            raise ValueError("Method generate_performance_data cannot be used if trace hasn't been activated.")

        for simulation_name, simulation in self.simulations.items():
            simulation_dir = os.path.join(output_dir, f"{simulation_name}_perf_log")
            if not os.path.exists(output_dir):
                os.mkdir(output_dir)
            if not os.path.exists(simulation_dir):
                os.mkdir(simulation_dir)
            simulation.tracer.generate_performance_graph(simulation_dir)
            simulation.tracer.generate_performance_tables(simulation_dir)

    def inflate(self, inflator_by_variable = None, period = None, target_by_variable = None):
        assert inflator_by_variable or target_by_variable
        assert period is not None
        inflator_by_variable = dict() if inflator_by_variable is None else inflator_by_variable
        target_by_variable = dict() if target_by_variable is None else target_by_variable
        self.inflator_by_variable = inflator_by_variable
        self.target_by_variable = target_by_variable

        for _, simulation in self.simulations.items():
            simulation.inflate(inflator_by_variable, period, target_by_variable)

    def init_from_data(self, calibration_kwargs = None, inflation_kwargs = None,
            rebuild_input_data = False, rebuild_kwargs = None, data = None, memory_config = None, use_marginal_tax_rate = False):
        """Initialise a survey scenario from data.

        Args:
          rebuild_input_data(bool):  Whether or not to clean, format and save data. Take a look at :func:`build_input_data`
          data(dict): Contains the data, or metadata needed to know where to find it.
          use_marginal_tax_rate(bool): True to go into marginal effective tax rate computation mode.
          calibration_kwargs(dict):  Calibration options (Default value = None)
          inflation_kwargs(dict): Inflations options (Default value = None)
          rebuild_input_data(bool): Whether to rebuild the data (Default value = False)
          rebuild_kwargs:  Rebuild options (Default value = None)
        """
        # When not ``None``, it'll try to get the data for *period*.
        if data is not None:
            data_year = data.get("data_year", self.period)

        # When ``True`` it'll assume it is raw data and do all that described supra.
        # When ``False``, it'll assume data is ready for consumption.
        if rebuild_input_data:
            if rebuild_kwargs is not None:
                self.build_input_data(year = data_year, **rebuild_kwargs)
            else:
                self.build_input_data(year = data_year)

        debug = self.debug
        trace = self.trace

        if use_marginal_tax_rate:
            for name, tax_benefit_system in self.tax_benefit_systems.items():
                assert self.varying_variable in tax_benefit_system.variables, f"Variable {self.varying_variable} is not present tax benefit system named {name}"

        # Inverting reform and baseline because we are more likely
        # to use baseline input in reform than the other way around
        self.simulations = dict()
        for simulation_name, _ in self.tax_benefit_systems.items():
            self.new_simulation(simulation_name, debug = debug, data = data, trace = trace, memory_config = memory_config)
            if use_marginal_tax_rate:
                self.new_simulation(simulation_name, debug = debug, data = data, trace = trace, memory_config = memory_config, marginal_tax_rate_only = True)

        if calibration_kwargs is not None:
            assert set(calibration_kwargs.keys()).issubset(set(
                ['target_margins_by_variable', 'parameters', 'total_population']))

        if inflation_kwargs is not None:
            assert set(inflation_kwargs.keys()).issubset(set(['inflator_by_variable', 'target_by_variable', 'period']))

        if calibration_kwargs:
            self.calibrate(**calibration_kwargs)

        if inflation_kwargs:
            self.inflate(**inflation_kwargs)

    def new_simulation(self, simulation_name, debug = False, trace = False, data = None, memory_config = None, marginal_tax_rate_only = False):
        tax_benefit_system = self.tax_benefit_systems[simulation_name]
        assert tax_benefit_system is not None

        period = periods.period(self.period)

        if 'custom_initialize' in dir(self):
            custom_initialize = (
                None
                if marginal_tax_rate_only
                else self.custom_initialize
                )
        else:
            custom_initialize = None

        data["collection"] = self.collection
        data["id_variable_by_entity_key"] = self.id_variable_by_entity_key
        data["role_variable_by_entity_key"] = self.role_variable_by_entity_key
        data["used_as_input_variables"] = self.used_as_input_variables

        simulation = Simulation.new_from_tax_benefit_system(
            tax_benefit_system = tax_benefit_system,
            debug = debug,
            trace = trace,
            data = data,
            memory_config = memory_config,
            period = period,
            custom_initialize = custom_initialize,
            )

        if marginal_tax_rate_only:
            self._apply_modification(simulation, period)
            if custom_initialize:
                custom_initialize(simulation)
            self.simulations[f"_modified_{simulation_name}"] = simulation
        else:
            self.simulations[simulation_name] = simulation

        simulation.weight_variable_by_entity = self.weight_variable_by_entity

        if self.period is not None:
            simulation.period = periods.period(self.period)

        return simulation

    def memory_usage(self):
        """Print memory usage."""
        for simulation_name, simulation in self.simulations.items():
            print(f"simulation : {simulation_name}")  # noqa analysis:ignore
            simulation.print_memory_usage()

    def neutralize_variables(self, tax_benefit_system):
        """Neutralizes input variables not in input dataframe and keep some crucial variables.

        Args:
          tax_benefit_system: The TaxBenefitSystem variables belongs to

        """
        for variable_name, variable in tax_benefit_system.variables.items():
            if variable.formulas:
                continue
            if self.used_as_input_variables and (variable_name in self.used_as_input_variables):
                continue
            if self.non_neutralizable_variables and (variable_name in self.non_neutralizable_variables):
                continue
            if self.weight_variable_by_entity and variable_name in list(self.weight_variable_by_entity.values()):
                continue

            tax_benefit_system.neutralize_variable(variable_name)

    def restore_simulations(self, directory, **kwargs):
        """Restores SurveyScenario's simulations.

        Args:
          directory: Directory to restore simulations from
          kwargs: Restoration options

        """
        assert os.path.exists(directory), "Cannot restore simulations from non existent directory"
        use_sub_directories = True if len(self.tax_benefit_systems) >= 2 else False

        self.simulations = dict()
        if use_sub_directories:
            for simulation_name, tax_benefit_system in self.tax_benefit_systems.items():
                simulation = restore_simulation(
                    os.path.join(directory, simulation_name),
                    tax_benefit_system,
                    **kwargs)
                simulation.id_variable_by_entity_key = self.id_variable_by_entity_key
                self.simulations[simulation_name] = simulation
        else:
            simulation = restore_simulation(directory, list(self.tax_benefit_systems.values())[0], **kwargs)
            simulation.id_variable_by_entity_key = self.id_variable_by_entity_key
            self.simulations["unique_simulation"] = simulation

    def set_input_data_frame(self, input_data_frame):
        """Set the input dataframe.

        Args:
          input_data_frame (pd.DataFrame): Input data frame

        """
        self.input_data_frame = input_data_frame

    def set_tax_benefit_systems(self, tax_benefit_systems: Dict[str, TaxBenefitSystem]):
        """
        Set the tax and benefit systems of the scenario.

        Args:
            tax_benefit_systems (Dict[str, TaxBenefitSystem]): The tax benefit systems
        """
        for tax_benefit_system in tax_benefit_systems.values():
            assert tax_benefit_system is not None
            if self.cache_blacklist is not None:
                tax_benefit_system.cache_blacklist = self.cache_blacklist
        #
        self.tax_benefit_systems = tax_benefit_systems

    def set_weight_variable_by_entity(self, weight_variable_by_entity = None):
        if weight_variable_by_entity is not None:
            self.weight_variable_by_entity = weight_variable_by_entity

        if self.simulations is not None:
            for simulation in self.simulations.values():
                simulation.set_weight_variable_by_entity(self.weight_variable_by_entity)

    def summarize_variable(self, variable = None, weighted = False, force_compute = False):
        """Print a summary of a variable including its memory usage for all the siulations.

        Args:
          variable(string): The variable being summarized
          weighted(bool): Whether the produced statistics should be weigthted or not
          force_compute(bool): Whether the computation of the variable should be forced

        Example:
            >>> from openfisca_survey_manager.tests.test_scenario import create_randomly_initialized_survey_scenario
            >>> survey_scenario = create_randomly_initialized_survey_scenario()
            >>> survey_scenario.summarize_variable(variable = "housing_occupancy_status", force_compute = True)
            <BLANKLINE>
            housing_occupancy_status: 1 periods * 5 cells * item size 2 (int16, default = HousingOccupancyStatus.tenant) = 10B
            Details:
            2017-01: owner = 0.00e+00 (0.0%), tenant = 5.00e+00 (100.0%), free_lodger = 0.00e+00 (0.0%), homeless = 0.00e+00 (0.0%).
            >>> survey_scenario.summarize_variable(variable = "rent", force_compute = True)
            <BLANKLINE>
            rent: 2 periods * 5 cells * item size 4 (float32, default = 0) = 40B
            Details:
            2017-01: mean = 562.385107421875, min = 156.01864624023438, max = 950.7142944335938, mass = 2.81e+03, default = 0.0%, median = 598.6585083007812
            2018-01: mean = 562.385107421875, min = 156.01864624023438, max = 950.7142944335938, mass = 2.81e+03, default = 0.0%, median = 598.6585083007812
            >>> survey_scenario.tax_benefit_systems["baseline"].neutralize_variable('age')
            >>> survey_scenario.summarize_variable(variable = "age")
            <BLANKLINE>
            age: neutralized variable (int64, default = 0)
        """
        for _simulation_name, simulation in self.simulations.items():
            simulation.summarize_variable(variable, weighted, force_compute)

    def _apply_modification(self, simulation, period):
        period = periods.period(period)
        varying_variable = self.varying_variable
        definition_period = simulation.tax_benefit_system.variables[varying_variable].definition_period

        def set_variable(varying_variable, varying_variable_value, period_):
            delta = self.variation_factor * varying_variable_value
            new_variable_value = varying_variable_value + delta
            simulation.delete_arrays(varying_variable, period_)
            simulation.set_input(varying_variable, period_, new_variable_value)

        if period.unit == definition_period:
            varying_variable_value = simulation.calculate(varying_variable, period = period)
            set_variable(varying_variable, varying_variable_value, period)

        elif (definition_period == MONTH) and (period.unit == YEAR and period.size_in_months == 12):
            varying_variable_value = simulation.calculate_add(varying_variable, period = period)
            for period_ in [periods.Period(('month', period.start.offset(month, 'month'), 1)) for month in range(12)]:
                set_variable(varying_variable, varying_variable_value / 12, period_)
        else:
            ValueError()





import numpy as np
import pandas as pd


from openfisca_core.model_api import Variable, YEAR
from openfisca_core.entities import build_entity
from openfisca_core.taxbenefitsystems import TaxBenefitSystem
from openfisca_survey_manager.scenarios.abstract_scenario import AbstractSurveyScenario
from openfisca_survey_manager.statshelpers import mark_weighted_percentiles
from openfisca_survey_manager.variables import quantile
from openfisca_survey_manager import default_config_files_directory

Individu = build_entity(
    key = "individu",
    plural = "individus",
    label = 'Individu',
    is_person = True,
    )

entities = [Individu]


class salaire(Variable):
    value_type = float
    entity = Individu
    label = "Salaire"
    definition_period = YEAR


class decile_salaire_from_quantile(Variable):
    entity = Individu
    value_type = int
    label = "Décile de salaire nouveau calcul"
    definition_period = YEAR
    formula = quantile(q = 10, variable = 'salaire')


class decile_salaire(Variable):
    value_type = int
    entity = Individu
    label = "Décile de salaire"
    definition_period = YEAR

    def formula(individu, period):
        salaire = individu('salaire', period)
        labels = np.arange(1, 11)
        weights = 1.0 * np.ones(shape = len(salaire))
        decile, _ = mark_weighted_percentiles(
            salaire,  # + np.random.uniform(size = len(salaire)) - 0.5,
            labels,
            weights,
            method = 2,
            return_quantiles = True,
            )
        return decile


class QuantileTestTaxBenefitSystem(TaxBenefitSystem):
    """PPDLand tax and benefit system"""
    CURRENCY = ""

    def __init__(self):
        super(QuantileTestTaxBenefitSystem, self).__init__(entities)
        for variable in [
                decile_salaire_from_quantile,
                decile_salaire,
                salaire,
                ]:
            self.add_variable(variable)


class QuantileTestSurveyScenario(AbstractSurveyScenario):
    def __init__(self, input_data_frame = None, tax_benefit_system = None,
            baseline_tax_benefit_system = None, period = None):
        super(QuantileTestSurveyScenario, self).__init__()
        assert input_data_frame is not None
        assert period is not None
        self.period = period
        if tax_benefit_system is None:
            tax_benefit_system = QuantileTestTaxBenefitSystem()

        tax_benefit_systems = (
            dict(
                reform = tax_benefit_system,
                baseline = baseline_tax_benefit_system
                )
            if baseline_tax_benefit_system
            else dict(baseline = tax_benefit_system)
            )

        self.set_tax_benefit_systems(tax_benefit_systems)

        self.used_as_input_variables = list(
            set(tax_benefit_system.variables.keys()).intersection(
                set(input_data_frame.columns)
                ))
        data = {
            'input_data_frame': input_data_frame,
            'config_files_directory': default_config_files_directory
            }
        self.init_from_data(data = data)


def create_input_dataframe(size = 9):
    """
    Create input dataframe with variable salaire and pension_retraite
    """
    np.random.seed(216)
    household_weight = 1.0
    number_of_indididual = size
    size = int(number_of_indididual / household_weight)
    salaire = np.linspace(0, 100, size)
    return pd.DataFrame({
        'salaire': salaire,
        })


def test_quantile():
    size = 10000
    input_data_frame = create_input_dataframe(size = size)
    survey_scenario = QuantileTestSurveyScenario(
        input_data_frame = input_data_frame,
        tax_benefit_system = QuantileTestTaxBenefitSystem(),
        period = 2017,
        )
    data = np.linspace(1, 11 - 1e-5, size)
    target = np.floor(data)
    result = survey_scenario.calculate_variable(
        variable = 'decile_salaire_from_quantile',
        period = '2017',
        simulation = "baseline",
        )
    assert all(
        (result == target) + (abs(result - target + 1) < .001)  # Finite size problem handling
        ), "{} != {}, \n{} , \n{},".format(
            result[result != target],
            target[result != target],
            data[result != target],
            abs(result - target + 1)[result != target],
            )

    # No reason that method coincides so close to the quantiles thresholds
    # assert all(survey_scenario.calculate_variable(
    #     variable = 'decile_salaire_from_quantile', period = '2017'
    #     ) == survey_scenario.calculate_variable(
    #         variable = 'decile_salaire', period = '2017'
    #         )
    #     )


if __name__ == '__main__':
    test_quantile()



#! /usr/bin/env python


"""Build or update a collection from raw surveys data."""


import argparse
import configparser
import datetime
import logging
import os
import pdb
import shutil
import sys


from openfisca_survey_manager.survey_collections import SurveyCollection
from openfisca_survey_manager.surveys import Survey
from openfisca_survey_manager import default_config_files_directory, openfisca_survey_manager_location

app_name = os.path.splitext(os.path.basename(__file__))[0]
log = logging.getLogger(app_name)


def add_survey_to_collection(survey_name = None, survey_collection = None, sas_files = None, stata_files = None, csv_files = None):
    if sas_files is None:
        sas_files = []
    if stata_files is None:
        stata_files = []
    if csv_files is None:
        csv_files = []

    assert survey_collection is not None
    overwrite = True
    label = survey_name

    for test_survey in survey_collection.surveys:
        if test_survey.name == survey_name:
            survey = survey_collection.get_survey(survey_name)
    if overwrite:
        survey = Survey(
            name = survey_name,
            label = label,
            csv_files = csv_files,
            sas_files = sas_files,
            stata_files = stata_files,
            survey_collection = survey_collection,
            )
    else:
        survey = survey_collection.get(survey_name)
        survey.label = label
        survey.informations.update({
            "csv_files": csv_files,
            "sas_files": sas_files,
            "stata_files": stata_files,
            })
    survey_collection.surveys = [
        kept_survey for kept_survey in survey_collection.surveys if kept_survey.name != survey_name
        ]
    survey_collection.surveys.append(survey)


def create_data_file_by_format(directory_path = None):
    """Browse subdirectories to extract stata and sas files."""
    stata_files = []
    sas_files = []
    csv_files = []

    for root, _subdirs, files in os.walk(directory_path):
        for file_name in files:
            file_path = os.path.join(root, file_name)
            if os.path.basename(file_name).endswith(".csv"):
                log.info("Found csv file {}".format(file_path))
                csv_files.append(file_path)
            if os.path.basename(file_name).endswith(".dta"):
                log.info("Found stata file {}".format(file_path))
                stata_files.append(file_path)
            if os.path.basename(file_name).endswith(".sas7bdat"):
                log.info("Found sas file {}".format(file_path))
                sas_files.append(file_path)
    return {'csv': csv_files, 'stata': stata_files, 'sas': sas_files}


def build_survey_collection(
        config_files_directory: str,
        collection_name = None,
        replace_metadata = False,
        replace_data = False,
        data_directory_path_by_survey_suffix = None,
        source_format = 'sas',
        ):

    assert collection_name is not None
    assert data_directory_path_by_survey_suffix is not None
    surveys_name = list(data_directory_path_by_survey_suffix.keys())
    assert surveys_name is not None, "A list of surveys to process is needed"

    if replace_metadata:
        survey_collection = SurveyCollection(
            name = collection_name, config_files_directory = config_files_directory)
    else:
        try:
            survey_collection = SurveyCollection.load(
                collection = collection_name, config_files_directory = config_files_directory)
        except configparser.NoOptionError:
            survey_collection = SurveyCollection(
                name = collection_name, config_files_directory = config_files_directory)

    for survey_suffix, data_directory_path in data_directory_path_by_survey_suffix.items():
        assert os.path.isdir(data_directory_path), '{} is not a valid directory path'.format(data_directory_path)

        data_file_by_format = create_data_file_by_format(data_directory_path)
        survey_name = '{}_{}'.format(collection_name, survey_suffix)
        add_survey_to_collection(
            survey_name = survey_name,
            survey_collection = survey_collection,
            csv_files = data_file_by_format.get('csv'),
            sas_files = data_file_by_format.get('sas'),
            stata_files = data_file_by_format.get('stata'),
            )

        valid_source_format = [
            _format for _format in list(data_file_by_format.keys())
            if data_file_by_format.get((_format))
            ]
        log.info("Valid source formats are: {}".format(valid_source_format))
        source_format = valid_source_format[0]
        log.info("Using the following format: {}".format(source_format))
        collections_directory = survey_collection.config.get('collections', 'collections_directory')
        if os.path.isdir(collections_directory) is False:
            log.info(
                "{} who should be the collections' directory does not exist. Creating directory.".format(
                    collections_directory))
            os.mkdir(collections_directory)
        collection_json_path = os.path.join(collections_directory, "{}.json".format(collection_name))
        survey_collection.dump(json_file_path = collection_json_path)
        surveys = [survey for survey in survey_collection.surveys if survey.name.endswith(str(survey_suffix))]
        survey_collection.fill_hdf(source_format = source_format, surveys = surveys, overwrite = replace_data)
    return survey_collection


def check_template_config_files(config_files_directory: str):
    raw_data_ini_path = os.path.join(config_files_directory, 'raw_data.ini')
    config_ini_path = os.path.join(config_files_directory, 'config.ini')
    raw_data_template_ini_path = os.path.join(config_files_directory, 'raw_data_template.ini')
    config_template_ini_path = os.path.join(config_files_directory, 'config_template.ini')

    if os.path.exists(config_files_directory):
        config_files_do_not_exist = not (os.path.exists(raw_data_ini_path) and os.path.exists(config_ini_path))
        templates_config_files_do_not_exist = not (
            os.path.exists(raw_data_template_ini_path) and os.path.exists(config_template_ini_path))

        if config_files_do_not_exist:
            if templates_config_files_do_not_exist:
                log.info("Creating configuration template files in {}".format(config_files_directory))
                template_files = [
                    'raw_data_template.ini', 'config_template.ini'
                    ]
                templates_config_files_directory = os.path.join(
                    openfisca_survey_manager_location,
                    'openfisca_survey_manager',
                    'config_files_templates'
                    )
                for template_file in template_files:
                    shutil.copy(
                        os.path.join(templates_config_files_directory, template_file),
                        os.path.join(config_files_directory, template_file),
                        )
            print("Rename and fill the template files in {}".format(config_files_directory))  # noqa analysis:ignore
            return False
    else:
        os.makedirs(config_files_directory)
        return check_template_config_files(config_files_directory)

    return True


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--collection', help = "name of collection to build or update", required = True)
    parser.add_argument('-d', '--replace-data', action = 'store_true', default = False,
        help = "erase existing survey data HDF5 file (instead of failing when HDF5 file already exists)")
    parser.add_argument('-m', '--replace-metadata', action = 'store_true', default = False,
        help = "erase existing collection metadata JSON file (instead of just adding new surveys)")
    parser.add_argument('-p', '--path', help = f'path to the config files directory (default = {default_config_files_directory})')
    parser.add_argument('-s', '--survey', help = 'name of survey to build or update (default = all)')
    parser.add_argument('-v', '--verbose', action = 'store_true', default = False, help = "increase output verbosity")
    args = parser.parse_args()
    logging.basicConfig(level = logging.DEBUG if args.verbose else logging.WARNING, stream = sys.stdout)

    if args.path:
        config_files_directory = args.path
    else:
        config_files_directory = default_config_files_directory

    if not check_template_config_files(config_files_directory = config_files_directory):
        return

    config_parser = configparser.ConfigParser()
    config_parser.read(os.path.join(config_files_directory, 'raw_data.ini'))
    assert config_parser.has_section(args.collection), '{} is an unkown collection. Please add a section to raw_data.ini configuration file'.format(
        args.collection)
    data_directory_path_by_survey_suffix = dict(config_parser.items(args.collection))
    if args.survey is not None:
        assert args.survey in data_directory_path_by_survey_suffix, 'Unknown survey data directory for {}'.format(
            args.collection)
        data_directory_path_by_survey_suffix = {
            args.survey: data_directory_path_by_survey_suffix[args.survey],
            }

    start_time = datetime.datetime.now()

    try:
        build_survey_collection(
            collection_name = args.collection,
            data_directory_path_by_survey_suffix = data_directory_path_by_survey_suffix,
            replace_metadata = args.replace_metadata,
            replace_data = args.replace_data,
            source_format = 'sas',
            config_files_directory = config_files_directory,
            )
    except Exception as e:
        log.info(e)
        pdb.post_mortem(sys.exc_info()[2])
        raise e

    log.info("The program has been executed in {}".format(datetime.datetime.now() - start_time))

    return 0


if __name__ == "__main__":
    sys.exit(main())





# OpenFisca -- A versatile microsimulation software
# By: OpenFisca Team <contact@openfisca.fr>
#
# Copyright (C) 2011, 2012, 2013, 2014, 2015 OpenFisca Team
import os
from pandas.testing import assert_frame_equal


from openfisca_survey_manager.read_sas import read_sas
from openfisca_survey_manager import openfisca_survey_manager_location


def test():
    sas_file_path = os.path.join(
        openfisca_survey_manager_location,
        'openfisca_survey_manager',
        'tests',
        'data_files',
        'help.sas7bdat',
        )
    data_frame = read_sas(sas_file_path, clean = False)
    data_frame_clean = read_sas(sas_file_path, clean = True)

    assert_frame_equal(data_frame, data_frame_clean)

    assert len(data_frame.columns) == 88
    assert len(data_frame) == 453
